from __future__ import annotations

from typing import Collection

import numpy as np

from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition

from .base import BaseResponse
from ..perturbation import DeltaKick, PerturbationLike, NoPerturbation
from ..density_matrices.frequency import (FrequencyDensityMatrices,
                                          FrequencyDensityMatricesFromDisk)
from ..density_matrices.time import (ConvolutionDensityMatricesFromDisk,
                                     ConvolutionDensityMatricesFromFrequency,
                                     ConvolutionDensityMatrices)
from ..utils import Logger
from ..typing import Array1D


class ResponseFromDensityMatrices(BaseResponse):

    """ Response from density matrices saved on disk.

    Parameters
    ----------
    pulserho_fmt
        Formatting string for the density matrices saved to disk.

        The formatting string should be a plain string containing variable
        placeholders within curly brackets `{}`. It should not be confused with
        a formatted string literal (f-string).

        Example:

        * pulserho_fmt =  `pulserho/t{time:09.1f}{tag}.npy`.

        Accepts variables

        * `{time}` - Time in units of as.
        * `{tag}` - Derivative tag, `''`, `'-Iomega'`, or `'-omega2'`.
        * `{pulsefreq}` - Pulse frequency in units of eV.
        * `{pulsefwhm}` - Pulse FWHM in units of fs.
    ksd
        KohnShamDecomposition object or file name.
    perturbation
        Perturbation that was present during time propagation.
    calc_size
        Size of the calculation communicator.
    """

    def __init__(self,
                 pulserho_fmt: str,
                 ksd: KohnShamDecomposition | str,
                 perturbation: PerturbationLike = None,
                 calc_size: int = 1):
        super().__init__(ksd=ksd,
                         perturbation=perturbation,
                         calc_size=calc_size)
        self.pulserho_fmt = pulserho_fmt

    def __str__(self) -> str:
        lines = [f'{self.__class__.__name__}']
        lines += [f'  ksd: {self.ksd.filename if self.ksd.filename is not None else "From calc"}']
        lines += [f'  pulserho_fmt: {self.pulserho_fmt}']
        lines += ['  perturbation:']
        lines += ['    ' + line for line in str(self.perturbation).split('\n')]
        return '\n'.join(lines)

    def _get_time_density_matrices(self,
                                   times: list[float] | Array1D[np.float64],
                                   pulses: Collection[PerturbationLike],
                                   derivative_order_s: list[int] = [0],
                                   real: bool = True,
                                   imag: bool = True,
                                   log: Logger | None = None,
                                   ) -> ConvolutionDensityMatrices:
        if (len(pulses) == 1 and list(pulses)[0] == self.perturbation) or isinstance(self.perturbation, NoPerturbation):
            # Yield the density matrices without performing convolution
            density_matrices = ConvolutionDensityMatricesFromDisk(
                     ksd=self.ksd,
                     pulserho_fmt=self.pulserho_fmt,
                     times=times,
                     pulses=pulses,
                     derivative_order_s=derivative_order_s,
                     real=real,
                     imag=imag,
                     log=log,
                     calc_size=self.calc_size)
        else:
            raise NotImplementedError('Pulse convolution of density matrices on disk is not implemented')

        return density_matrices

    def _get_frequency_density_matrices(self,
                                        frequencies: list[float] | Array1D[np.float64],
                                        frequency_broadening: float = 0,
                                        real: bool = True,
                                        imag: bool = True,
                                        log: Logger | None = None,
                                        ) -> FrequencyDensityMatrices:
        raise NotImplementedError('Fourier transformation of density matrices on disk is not implemented')


class ResponseFromFourierTransform(BaseResponse):

    """ Response from Fourier transform of density matrices save on disk.

    Parameters
    ----------
    frho_fmt
        Formatting string for the density matrices
        in frequency space saved to disk.

        The formatting string should be a plain string containing variable
        placeholders within curly brackets `{}`. It should not be confused with
        a formatted string literal (f-string).

        Example:

        * frho_fmt = `frho/w{freq:05.2f}-{reim}.npy`.

        Accepts variables:

        * `{freq}` - Frequency in units of eV.
        * `{reim}` - `'Re'` or `'Im'` for Fourier transform of real/imaginary
          part of density matrix.
    ksd
        KohnShamDecomposition object or file name.
    perturbation
        Perturbation that was present during time propagation.
    calc_size
        Size of the calculation communicator.
    """

    def __init__(self,
                 frho_fmt: str,
                 ksd: KohnShamDecomposition | str,
                 perturbation: PerturbationLike = None,
                 calc_size: int = 1):
        super().__init__(ksd=ksd,
                         perturbation=perturbation,
                         calc_size=calc_size)

        if not isinstance(self.perturbation, (DeltaKick, NoPerturbation)):
            raise NotImplementedError('Only delta kick implemented')

        self.frho_fmt = frho_fmt

    def __str__(self) -> str:
        lines = [f'{self.__class__.__name__}']
        lines += [f'  ksd: {self.ksd.filename if self.ksd.filename is not None else "From calc"}']
        lines += [f'  frho_fmt: {self.frho_fmt}']
        lines += ['  perturbation:']
        lines += ['    ' + line for line in str(self.perturbation).split('\n')]
        return '\n'.join(lines)

    def _get_time_density_matrices(self,
                                   times: list[float] | Array1D[np.float64],
                                   pulses: Collection[PerturbationLike],
                                   derivative_order_s: list[int] = [0],
                                   real: bool = True,
                                   imag: bool = True,
                                   log: Logger | None = None,
                                   ) -> ConvolutionDensityMatrices:
        if not isinstance(self.perturbation, DeltaKick):
            raise NotImplementedError('Only delta kick implemented')

        density_matrices = ConvolutionDensityMatricesFromFrequency(
            ksd=self.ksd,
            frho_fmt=self.frho_fmt,
            perturbation=self.perturbation,
            pulses=pulses,
            times=times,
            derivative_order_s=derivative_order_s,
            real=real,
            imag=imag,
            log=log,
            calc_size=self.calc_size)

        return density_matrices

    def _get_frequency_density_matrices(self,
                                        frequencies: list[float] | Array1D[np.float64],
                                        frequency_broadening: float = 0,
                                        real: bool = True,
                                        imag: bool = True,
                                        log: Logger | None = None,
                                        ) -> FrequencyDensityMatrices:
        if frequency_broadening > 0:
            raise NotImplementedError(f'Frequency broadening not implemented for {self.__class__.__name__}')

        density_matrices = FrequencyDensityMatricesFromDisk(
            ksd=self.ksd,
            frho_fmt=self.frho_fmt,
            perturbation=self.perturbation,
            frequencies=frequencies,
            real=real,
            imag=imag,
            log=log,
            calc_size=self.calc_size)

        return density_matrices



from __future__ import annotations

from typing import Generator
from pathlib import Path

import numpy as np

from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition
from gpaw.tddft.units import au_to_eV, eV_to_au
from gpaw.mpi import world

from .distributed.frequency import FourierTransformer
from .density_matrix import DensityMatrix
from .readers.numpy import FrequencyDensityMatrixReader
from .base import BaseDensityMatrices, WorkMetadata
from ..perturbation import create_perturbation, PerturbationLike, DeltaKick, NoPerturbation
from ..utils import Logger, two_communicator_sizes
from ..utils.memory import MemoryEstimate
from ..typing import Array1D


class FrequencyDensityMatrixMetadata(WorkMetadata):

    """ Metadata to the density matrices.

    Parameters
    ----------
    density_matrices
        Parent of this object.
    globalw
        Frequency index.
    localw
        Frequency index on this rank.
    globalr
        Real/imaginary index.
    localr
        Real/imaginary index on this rank.
    """

    density_matrices: FrequencyDensityMatrices
    globalw: int
    localw: int
    globalr: int
    localr: int

    def __new__(cls,
                density_matrices: FrequencyDensityMatrices,
                globalw: int,
                globalr: int,
                localw: int,
                localr: int):
        self = WorkMetadata.__new__(cls, density_matrices=density_matrices)
        self.globalw = globalw
        self.globalr = globalr
        self.localw = localw
        self.localr = localr
        return self

    @property
    def global_indices(self):
        return (self.globalw, self.globalr)

    @property
    def freq(self) -> float:
        """ Frequency in units of eV. """
        return self.density_matrices.frequencies[self.globalw]

    @property
    def reim(self) -> str:
        """ Returns real/imaginary tag ``'Re'`` or ``'Im'``.

        The tag corresponds to the Fourier transform of the real
        or imaginary part of the density matrix.
        """
        return self.density_matrices.reim[self.globalr]

    @property
    def desc(self) -> str:
        return f'{self.reim} @ Freq. {self.freq:.3f}eV'

    @property
    def real(self) -> bool:
        return self.reim == 'Re'

    @property
    def imag(self) -> bool:
        return not self.real


class FrequencyDensityMatrices(BaseDensityMatrices[FrequencyDensityMatrixMetadata]):

    """
    Collection of density matrices in the Kohn-Sham basis in the frequency
    domain, for different frequencies.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    frequencies
        Compute density matrices for these frequencies (or as close to them as possible). In units of eV.
    frequency_broadening
        Gaussian broadening width in units of eV. Default (0) is no broadening.
    real
        Calculate the Fourier transform of the real part of the density matrix.
    imag
        Calculate the Fourier transform of the imaginary part of the density matrix.
    calc_size
        Size of the calculation communicator.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 frequencies: list[float] | Array1D[np.float64],
                 frequency_broadening: float = 0,
                 real: bool = True,
                 imag: bool = True,
                 calc_size: int = 1,
                 log: Logger | None = None):
        self._freq_w = np.array(frequencies)
        self._frequency_broadening = frequency_broadening

        super().__init__(ksd=ksd,
                         real=real, imag=imag,
                         calc_size=calc_size, log=log)

    @property
    def frequencies(self) -> Array1D[np.float64]:
        """ Frequencies in units of eV. """
        return self._freq_w  # type: ignore

    @property
    def frequency_broadening(self) -> float:
        """ Gaussian broadening width for frequencies in units of eV. """
        return self._frequency_broadening

    def work_loop(self,
                  rank: int) -> Generator[FrequencyDensityMatrixMetadata | None, None, None]:
        nw = len(self.frequencies)
        nwperrank = (nw + self.loop_comm.size - 1) // self.loop_comm.size

        for localw in range(nwperrank):
            globalw = rank + localw * self.loop_comm.size
            for r in range(len(self.reim)):
                if globalw < nw:
                    yield FrequencyDensityMatrixMetadata(density_matrices=self, globalw=globalw,
                                                         localw=localw, globalr=r, localr=r)
                else:
                    yield None

    def write_to_disk(self,
                      frho_fmt: str):
        """ Calculate the density matrices and save to disk.

        Parameters
        ----------
        frho_fmt
            Formatting string for the density matrices saved to disk.

            The formatting string should be a plain string containing variable
            placeholders within curly brackets ``{}``. It should not be confused with
            a formatted string literal (f-string).

            Example:

             * frho_fmt = ``frho/w{freq:05.2f}-{reim}.npy``.

            Accepts variables:

             * ``{freq}`` - Frequency in units of eV.
             * ``{reim}`` - ``'Re'`` or ``'Im'`` for Fourier transform of real/imaginary
               part of density matrix.
        """
        nlocaltot = len(self.local_work_plan)

        # Iterate over density matrices on all ranks
        for ndone, (work, dm) in enumerate(self, 1):
            avg = self.log.elapsed('read')/ndone
            estrem = avg * (nlocaltot - ndone)
            self.log(f'Obtained density matrix {ndone:4.0f}/{nlocaltot:4.0f} on this rank. '
                     f'Avg: {avg:10.3f}s, estimated remaining: {estrem:10.3f}s', who='Writer', flush=True)
            fname_kw = dict(freq=work.freq, reim=work.reim)
            fpath = Path(frho_fmt.format(**fname_kw))
            rho_ia = dm.rho_ia
            if self.calc_comm.rank == 0:
                assert isinstance(rho_ia, np.ndarray)
                fpath.parent.mkdir(parents=True, exist_ok=True)
                if fpath.suffix == '.npz':
                    # Write numpy archive
                    np.savez(str(fpath), rho_ia=rho_ia)
                else:
                    # Save numpy binary file
                    np.save(str(fpath), rho_ia)
                self.log(f'Written {fpath}', who='Writer', flush=True)
        world.barrier()


class FrequencyDensityMatricesFromDisk(FrequencyDensityMatrices):

    """
    Collection of density matrices in the Kohn-Sham basis in the frequency
    domain, for different frequencies. Read from disk.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    frho_fmt
        Formatting string for the density matrices
        in frequency space saved to disk.

        The formatting string should be a plain string containing variable
        placeholders within curly brackets ``{}``. It should not be confused with
        a formatted string literal (f-string).

        Example:

         * frho_fmt = ``frho/w{freq:05.2f}-{reim}.npy``.

        Accepts variables:

         * ``{freq}`` - Frequency in units of eV.
         * ``{reim}`` - ``'Re'`` or ``'Im'`` for Fourier transform of real/imaginary
           part of density matrix.
    perturbation
        The perturbation which the density matrices are a response to.
    frequencies
        Compute density matrices for these frequencies (or as close to them as possible). In units of eV.
    real
        Calculate the Fourier transform of the real part of the density matrix
    imag
        Calculate the Fourier transform of the imaginary part of the density matrix.
    calc_size
        Size of the calculation communicator.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 frho_fmt: str,
                 perturbation: PerturbationLike,
                 frequencies: list[float] | Array1D[np.float64],
                 real: bool = True,
                 imag: bool = True,
                 calc_size: int = 1,
                 log: Logger | None = None):
        reader = FrequencyDensityMatrixReader(frho_fmt=frho_fmt,
                                              ksd=ksd,
                                              filter_frequencies=frequencies,
                                              real=real, imag=imag,
                                              log=log)
        self.reader = reader
        super().__init__(ksd=reader.ksd, frequencies=reader.frequencies,
                         real=real, imag=imag,
                         calc_size=calc_size,
                         log=reader.log)
        self.perturbation = create_perturbation(perturbation)
        assert isinstance(self.perturbation, (DeltaKick, NoPerturbation))

    def __str__(self) -> str:
        return str(self.reader)

    def __iter__(self) -> Generator[tuple[FrequencyDensityMatrixMetadata, DensityMatrix], None, None]:
        for work in self.local_work_plan:
            self.log.start('read')
            rho = self.reader.read(work.freq, work.reim == 'Re')
            if isinstance(self.perturbation, DeltaKick):
                # Only delta kick supported at this point for normalization
                rho /= self.perturbation.strength
            matrices = {0: rho if self.calc_comm.rank == 0 else None}
            dm = DensityMatrix(ksd=self.ksd, matrices=matrices)

            yield work, dm


class FrequencyDensityMatricesFromWaveFunctions(FrequencyDensityMatrices):

    """
    Collection of density matrices in the Kohn-Sham basis in the frequency
    domain, for different frequencies. Obtained from the time-dependent wave functions file,
    which is read from disk.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    wfs_fname
        File name of the time-dependent wave functions file.
    perturbation
        The perturbation which the density matrices are a response to.
    frequencies
        Compute density matrices for these frequencies (or as close to them as possible). In units of eV.
    frequency_broadening
        Gaussian broadening width in units of eV. Default (0) is no broadening.
    real
        Calculate the real part of density matrices.
    imag
        Calculate the imaginary part of density matrices.
    calc_size
        Size of the calculation communicator.
    stridet
        Skip this many steps when reading the time-dependent wave functions file.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 wfs_fname: str,
                 perturbation: PerturbationLike,
                 frequencies: list[float] | Array1D[np.float64],
                 frequency_broadening: float = 0,
                 real: bool = True,
                 imag: bool = True,
                 log: Logger | None = None,
                 calc_size: int = 1,
                 stridet: int = 1):
        _, calc_size = two_communicator_sizes(-1, calc_size)
        # The calc_comm rank 0's are world ranks 0, with a spacing of calc_size
        result_on_ranks = list(range(0, world.size, calc_size))

        rho_nn_fft = FourierTransformer.from_parameters(
                wfs_fname, ksd,
                perturbation=perturbation,
                yield_re=real,
                yield_im=imag,
                filter_frequencies=np.array(frequencies) * eV_to_au,
                frequency_broadening=frequency_broadening * eV_to_au,
                stridet=stridet,
                result_on_ranks=result_on_ranks,
                log=log)
        self.rho_nn_fft = rho_nn_fft
        super().__init__(ksd=rho_nn_fft.ksd, frequencies=rho_nn_fft.freq_w * au_to_eV,
                         frequency_broadening=frequency_broadening,
                         real=real, imag=imag,
                         calc_size=calc_size,
                         log=rho_nn_fft.log)

    def __str__(self) -> str:
        lines = []
        lines.append('Response from time-dependent wave functions')
        lines.append('')
        lines += ['' + line for line in str(self.rho_nn_fft.rho_nn_reader.rho_wfs_reader).split('\n')]
        lines.append('')
        lines += ['' + line for line in str(self.rho_nn_fft.rho_nn_reader).split('\n')]
        lines.append('')
        lines += ['' + line for line in str(self.rho_nn_fft).split('\n')]

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        return self.rho_nn_fft.get_memory_estimate()

    @property
    def myw(self) -> list[int]:
        """ List of indices corresponding to the frequency indices on held on this rank. """
        return self.rho_nn_fft.my_work()

    def __iter__(self) -> Generator[tuple[FrequencyDensityMatrixMetadata, DensityMatrix], None, None]:
        parameters = self.rho_nn_fft.rho_nn_reader._parameters
        flt = (slice(parameters.n1size), slice(parameters.n2size))

        dist_buffer = self.rho_nn_fft.dist_buffer  # Perform the redistribution
        self.ksd.distribute(self.calc_comm)

        for work in self.local_work_plan:
            if self.calc_comm.rank == 0:
                assert self.myw[work.localw] == work.globalw

            if self.calc_comm.rank == 0:
                # Buffer shape is i, a, frequencies
                rho_ia = dist_buffer._get_data(work.reim == 'Re', 0)[flt + (work.localw, )]
            matrices = {0: rho_ia if self.calc_comm.rank == 0 else None}
            dm = DensityMatrix(ksd=self.ksd, matrices=matrices, comm=self.calc_comm)

            yield work, dm

    def parallel_prepare(self):
        self.rho_nn_fft.dist_buffer  # Perform the redistribution



from __future__ import annotations

import os
import sys
import time

import numpy as np

import ase
import gpaw
from gpaw.mpi import world
from gpaw.tddft.units import au_to_eV, au_to_as

from .. import __version__
from ..typing import Communicator

ascii_icon = r"""
                ###     #####
               ###############
         ########  #####     ##
     ###########            ###
  #######                  ###
 ####                      ##
 #                          ##
 #                           ##
 #                            ##
 #                             ##
 ##                             ###
  ####                           ###
   ##########                     ##
      ########                   ###
            ###                ####
            ##               ####
            ##           ######
            #################
"""


ascii_logo = r"""
      _               _            _
 _ __| |__   ___   __| | ___ _ __ | |_
| '__| '_ \ / _ \ / _` |/ _ \ '_ \| __|
| |  | | | | (_) | (_| |  __/ | | | |_
|_|  |_| |_|\___/ \__,_|\___|_| |_|\__|
"""


class Logger:

    """ Logger

    Parameters
    ----------
    t0
        Start time (default is current time).
    """
    _t0: float
    _starttimes: dict[str, float]

    def __init__(self,
                 t0: float | None = None):
        self._starttimes = dict()
        if t0 is None:
            self._t0 = time.time()
        else:
            assert isinstance(t0, float)
            self._t0 = t0
        self._time_of_last_log = self._t0

    @property
    def t0(self) -> float:
        return self._t0

    @t0.setter
    def t0(self,
           value: float | None):
        if value is None:
            self._t0 = time.time()
            return
        assert isinstance(value, float)
        self._t0 = value

    def __getitem__(self, key) -> float:
        return self._starttimes.get(key, self.t0)

    def __call__(self,
                 *args,
                 who: str | None = None,
                 rank: int | None = None,
                 if_elapsed: float = 0,
                 comm: Communicator | None = None,
                 **kwargs):
        """ Log message.

        Parameters
        ----------
        rank
            Only log if rank is :attr:`rank`. ``None`` to always log.
        who
            Sender of the message.
        comm
            Communicator. If included, rank and size is included in the message.
        if_elapsed
            Only log if :attr:`if_elapsed` seconds have passed since last logged message.
        """

        myrank = world.rank if comm is None else comm.rank
        if rank is not None and myrank != rank:
            return
        if time.time() < self._time_of_last_log + if_elapsed:
            return
        if comm is not None and comm.size > 1:
            commstr = f'{comm.rank:04.0f}/{comm.size:04.0f}'
            who = commstr if who is None else f'{who} {commstr}'
        _args = list(args)
        if who is not None:
            _args.insert(0, f'[{who}]')
        return self.log(*_args, **kwargs)

    def __str__(self) -> str:
        s = f'{self.__class__.__name__} t0: {self.t0}'
        return s

    def log(self, *args, **kwargs):
        """ Log message, prepending a timestamp. """
        self._time_of_last_log = time.time()
        hh, rem = divmod(self._time_of_last_log - self.t0, 3600)
        mm, ss = divmod(rem, 60)
        timestr = f'[{hh:02.0f}:{mm:02.0f}:{ss:04.1f}]'
        print(f'{timestr}', *args, **kwargs)

    def start(self, key):
        self._starttimes[key] = time.time()

    def elapsed(self, key) -> float:
        return time.time() - self[key]

    def startup_message(self):
        """ Print a start up message. """
        if world.rank != 0:
            return

        # Piece together logotype and version number
        logo_lines = ascii_logo.split('\n')
        width = max(len(line) for line in logo_lines) + 2
        i = -2
        logo_lines[i] += (width - len(logo_lines[i])) * ' '  # Pad to width
        logo_lines[i] += __version__

        # Piece together icon and logotype
        lines = ascii_icon.split('\n')
        width = max(len(line) for line in lines)

        for i, logoline in enumerate(logo_lines, start=3):
            line = lines[i]
            line += (width - len(line)) * ' '  # Pad to same length
            lines[i] = line + logoline

        print('\n'.join(lines))
        print('Date:  ', time.asctime())
        print('CWD:   ', os.getcwd())
        print('cores: ', world.size)
        print('Python: {}.{}.{}'.format(*sys.version_info[:3]))
        print(f'numpy:  {os.path.dirname(np.__file__)} (version {np.version.version})')
        print(f'ASE:    {os.path.dirname(ase.__file__)} (version {ase.__version__})')
        print(f'GPAW:   {os.path.dirname(gpaw.__file__)} (version {gpaw.__version__})')
        print(flush=True)


class NoLogger(Logger):

    def __str__(self) -> str:
        return self.__class__.__name__

    def log(self, *args, **kwargs):
        pass


def format_times(times: np.typing.ArrayLike,
                 units: str = 'as') -> str:
    """ Write a short list of times for pretty priting.

    Parameters
    ----------
    times
        List of times in units of :attr:`units`.
    units
        Units of the supplied times.

         * ``au`` - atomic units
         * ``as`` - attoseconds

    Returns
    -------
    Formatted list of times in units of as.
    """
    times = np.array(times)
    if units == 'au':
        times *= au_to_as
    elif units != 'as':
        raise ValueError(f'Unknown units {units}. Must be "au" or "as".')
    if len(times) < 5:
        # Print all times
        timesstrings = [f'{time:.1f}' for time in times]
    else:
        timesstrings = [f'{time:.1f}' for time in times[[0, 1, 2, -1]]]
        timesstrings.insert(-1, '...')
    timesstrings[-1] += ' as'
    return ', '.join(timesstrings)


def format_frequencies(frequencies: np.typing.ArrayLike,
                       units: str = 'eV') -> str:
    """ Write a short list of frequencies for pretty priting.

    Parameters
    ----------
    frequencies
        List of frequencies in units of :attr:`units`.
    units
        Units of the supplied frequencies.

         * ``au`` - atomic units
         * ``eV`` - electron volts

    Returns
    -------
    Formatted list of times in units of as.
    """
    frequencies = np.array(frequencies)
    if units == 'au':
        frequencies *= au_to_eV
    elif units != 'eV':
        raise ValueError(f'Unknown units {units}. Must be "au" or "eV".')

    if len(frequencies) < 5:
        # Print all frequencies
        freqsstrings = [f'{freq:.1f}' for freq in frequencies]
    else:
        freqsstrings = [f'{freq:.1f}' for freq in frequencies[[0, 1, 2, -1]]]
        freqsstrings.insert(-1, '...')
    freqsstrings[-1] += ' eV'
    return ', '.join(freqsstrings)



from __future__ import annotations

from typing import Generator, Collection

from pathlib import Path
import numpy as np
from numpy.typing import NDArray

from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition
from gpaw.tddft.units import as_to_au, au_to_as, eV_to_au, au_to_eV
from gpaw.mpi import world

from .buffer import DensityMatrixBuffer
from .distributed.pulse import PulseConvolver
from .readers.gpaw import KohnShamRhoWfsReader
from .readers.numpy import FrequencyDensityMatrixReader, TimeDensityMatrixReader
from .density_matrix import DensityMatrix
from .base import BaseDensityMatrices, WorkMetadata
from ..perturbation import create_perturbation, Perturbation, DeltaKick, PerturbationLike, PulsePerturbation
from ..utils import Logger, two_communicator_sizes, get_gaussian_pulse_values
from ..utils.logging import format_times
from ..utils.memory import MemoryEstimate
from ..typing import Array1D, Communicator


class ConvolutionDensityMatrixMetadata(WorkMetadata):
    """ Metadata to the density matrices.

    Properties
    ----------
    density_matrices
        Parent of this object.
    globalt
        Time index.
    localt
        Time index on this rank.
    globalp
        Pulse index.
    localp
        Pulse index on this rank.
    """
    density_matrices: ConvolutionDensityMatrices
    globalt: int
    globalp: int
    localt: int
    localp: int

    def __new__(cls,
                density_matrices: ConvolutionDensityMatrices,
                globalt: int,
                globalp: int,
                localt: int,
                localp: int):
        self = WorkMetadata.__new__(cls, density_matrices=density_matrices)
        self.globalt = globalt
        self.globalp = globalp
        self.localt = localt
        self.localp = localp
        return self

    @property
    def global_indices(self):
        return (self.globalp, self.globalt)

    @property
    def time(self) -> float:
        """ Simulation time in units of as. """
        return self.density_matrices.times[self.globalt]

    @property
    def pulse(self) -> Perturbation:
        """ Pulse. """
        return self.density_matrices.pulses[self.globalp]

    @property
    def desc(self) -> str:
        timestr = f'{self.time:.1f}as'
        if len(self.density_matrices.pulses) == 1:
            return timestr

        d = self.pulse.todict()
        if d['name'] == 'GaussianPulse':
            pulsestr = f'{d["frequency"]:.1f}eV'
        else:
            pulsestr = f'#{self.globalp}'
        return f'{timestr} @ Pulse {pulsestr}'


class ConvolutionDensityMatrices(BaseDensityMatrices[ConvolutionDensityMatrixMetadata]):

    """
    Collection of density matrices in the Kohn-Sham basis for different times,
    after convolution with various pulses.

    Plain density matrices and/or derivatives thereof may be represented.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    pulses
        Convolute the density matrices with these pulses.
    times
        Compute density matrices for these times (or as close to them as possible). In units of as.
    derivative_order_s
        Compute density matrix derivatives of the following orders.
        0 for plain density matrix and positive integers for derivatives.
    real
        Calculate the real part of density matrices.
    imag
        Calculate the imaginary part of density matrices.
    calc_size
        Size of the calculation communicator.
    """

    _derivative_order_s: list[int]

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 pulses: Collection[PerturbationLike],
                 times: list[float] | NDArray[np.float64],
                 derivative_order_s: list[int] = [0],
                 real: bool = True,
                 imag: bool = True,
                 calc_size: int = 1,
                 log: Logger | None = None):
        self._time_t = np.array(times)
        self._pulses = [create_perturbation(pulse) for pulse in pulses]
        self._derivative_order_s = derivative_order_s

        super().__init__(ksd=ksd,
                         real=real, imag=imag,
                         calc_size=calc_size, log=log)

        tag_s = ['', '-Iomega', '-omega2']
        return_keys_by_order = {0: 'pulserho', 1: 'pulsedrho', 2: 'pulseddrho'}
        self.keys_by_order = {0: 'rho_ia', 1: 'drho_ia', 2: 'ddrho_ia'}
        if 1 not in derivative_order_s:
            tag_s.remove('-Iomega')
            return_keys_by_order.pop(1)
        if 2 not in derivative_order_s:
            tag_s.remove('-omega2')
            return_keys_by_order.pop(2)
        self.tag_s = tag_s
        self.return_keys_by_order = return_keys_by_order

    @property
    def times(self) -> Array1D[np.float64]:
        """ Simulation time in units of as. """
        return self._time_t  # type: ignore

    @property
    def nt(self) -> int:
        """ Number of times. """
        return len(self.times)

    @property
    def pulses(self) -> list[Perturbation]:
        """ Pulses with which density matrices are convoluted. """
        return self._pulses

    @property
    def derivative_order_s(self) -> list[int]:
        """
        List of orders of the density matrix derivatives to be computed.
        `0` for plain density matrix and positive integers for derivatives.
        """
        return self._derivative_order_s

    def work_loop(self,
                  rank: int) -> Generator[ConvolutionDensityMatrixMetadata | None, None, None]:
        nt = len(self.times)
        ntperrank = (nt + self.loop_comm.size - 1) // self.loop_comm.size

        # Do convolution pulse-by-pulse, time-by-time
        for p, pulse in enumerate(self.pulses):
            # Determine which times to compute on this loop_comm rank for good load balancing
            shift = (p * nt + rank) % self.loop_comm.size
            for localt in range(ntperrank):
                globalt = shift + localt * self.loop_comm.size
                if globalt < nt:
                    yield ConvolutionDensityMatrixMetadata(density_matrices=self, globalt=globalt,
                                                           localt=localt, globalp=p, localp=p)
                else:
                    yield None

    def write_to_disk(self,
                      pulserho_fmt: str):
        """ Calculate the density matrices and save to disk.

        Parameters
        ----------
        pulserho_fmt
            Formatting string for the density matrices saved to disk.

            The formatting string should be a plain string containing variable
            placeholders within curly brackets `{}`. It should not be confused with
            a formatted string literal (f-string).

            Example:

            * pulserho_fmt =  `pulserho/t{time:09.1f}{tag}.npy`.

            Accepts variables

            * `{time}` - Time in units of as.
            * `{tag}` - Derivative tag, `''`, `'-Iomega'`, or `'-omega2'`.
            * `{pulsefreq}` - Pulse frequency in units of eV.
            * `{pulsefwhm}` - Pulse FWHM in units of fs.
        """
        nlocaltot = len(self.local_work_plan)

        tags_keys = [(tag, key) for s, (tag, key) in enumerate(
            [('', 'rho_ia'), ('-Iomega', 'drho_ia'), ('-omega2', 'ddrho_ia')]) if s in self.derivative_order_s]

        # Iterate over density matrices on all ranks
        for ndone, (work, dm) in enumerate(self, 1):
            avg = self.log.elapsed('read')/ndone
            estrem = avg * (nlocaltot - ndone)
            self.log(f'Obtained density matrix {ndone:4.0f}/{nlocaltot:4.0f} on this rank. '
                     f'Avg: {avg:10.3f}s, estimated remaining: {estrem:10.3f}s', who='Writer', flush=True)
            for tag, key in tags_keys:
                fname_kw = dict(time=work.time, tag=tag, **get_gaussian_pulse_values(work.pulse))
                fpath = Path(pulserho_fmt.format(**fname_kw))
                rho_ia = getattr(dm, key)
                if self.calc_comm.rank == 0:
                    assert isinstance(rho_ia, np.ndarray)
                    fpath.parent.mkdir(parents=True, exist_ok=True)
                    if fpath.suffix == '.npz':
                        # Write numpy archive
                        np.savez(str(fpath), rho_ia=rho_ia)
                    else:
                        # Save numpy binary file
                        np.save(str(fpath), rho_ia)
                    self.log(f'Written {fpath}', who='Writer', flush=True)
        world.barrier()


class TimeDensityMatricesFromWaveFunctions(ConvolutionDensityMatrices):

    """
    Collection of density matrices in the Kohn-Sham basis for different times,
    obtained by reading the time-dependent wave functions file.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    wfs_fname
        File name of the time-dependent wave functions file.
    times
        Compute density matrices for these times (or as close to them as possible). In units of as.
    real
        Calculate the real part of density matrices.
    imag
        Calculate the imaginary part of density matrices.
    calc_size
        Size of the calculation communicator.
    stridet
        Skip this many steps when reading the time-dependent wave functions file.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 wfs_fname: str,
                 times: list[float] | Array1D[np.float64],
                 real: bool = True,
                 imag: bool = True,
                 calc_size: int = 1,
                 log: Logger | None = None,
                 stridet: int = 1):
        rho_nn_direct = KohnShamRhoWfsReader(wfs_fname=wfs_fname, ksd=ksd,
                                             yield_re=real, yield_im=imag,
                                             filter_times=np.array(times) * as_to_au,  # type: ignore
                                             stridet=stridet, log=log)
        self.rho_nn_direct = rho_nn_direct

        super().__init__(ksd=rho_nn_direct.ksd,
                         times=rho_nn_direct.time_t * au_to_as,
                         real=real,
                         imag=imag,
                         pulses=[None],
                         calc_size=1,
                         log=rho_nn_direct.log)

        imin, imax, amin, amax = self.ksd.ialims()

        # Read density matrices corresponding to ksd ialims
        self._n1slice = slice(imin, imax + 1)
        self._n2slice = slice(amin, amax + 1)

    def __str__(self) -> str:
        lines = ['Response from time-dependent wave functions']

        lines.append('')
        lines += ['  ' + line for line in str(self.rho_nn_direct).split('\n')]
        shape = (self._n1slice.stop - self._n1slice.start,
                 self._n2slice.stop - self._n2slice.start,)
        if 'Re' in self.reim and 'Im' in self.reim:
            reim_desc = 'real and imaginary parts'
        elif 'Re' in self.reim:
            reim_desc = 'real parts'
        else:
            reim_desc = 'imaginary parts'

        lines.append('')
        lines.append(f'Density matrix shape {shape}')
        lines.append(f'reading {reim_desc}')

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        shape = (self._n1slice.stop - self._n1slice.start,
                 self._n2slice.stop - self._n2slice.start,)
        narrays = 2 if 'Re' in self.reim and 'Im' in self.reim else 1
        for t_r in self.rho_nn_direct.work_loop_by_ranks():
            nreading = len(t_r)
            break
        memory_estimate = MemoryEstimate()
        memory_estimate.add_child('Time-dependent wave functions reader',
                                  self.rho_nn_direct.get_memory_estimate())
        memory_estimate.add_key('Density matrix', (shape) + (narrays, ),
                                on_num_ranks=nreading)

        return memory_estimate

    def __iter__(self) -> Generator[tuple[ConvolutionDensityMatrixMetadata, DensityMatrix], None, None]:
        assert self.calc_comm.size == 1  # TODO

        for work, dm_buffer in zip(self.work_loop(self.loop_comm.rank),
                                   self.rho_nn_direct.iread(0, 0, self._n1slice, self._n2slice)):
            assert work is not None
            if 'Re' in self.reim:
                Rerho_ia = dm_buffer._get_real(0)
            if 'Im' in self.reim:
                Imrho_ia = dm_buffer._get_imag(0)
            if 'Re' in self.reim and 'Im' in self.reim:
                # Complex result
                # Compared to numpy, we use another convention, hence the minus sign
                rho_ia = Rerho_ia - 1j * Imrho_ia
            elif 'Re' in self.reim:
                # Real result
                rho_ia = Rerho_ia
            else:
                rho_ia = -1j * Imrho_ia
            matrices: dict[int, NDArray[np.complex128] | None] = {0: rho_ia}
            dm = DensityMatrix(ksd=self.ksd, matrices=matrices, comm=self.calc_comm)

            yield work, dm

    def parallel_prepare(self):
        self.rho_nn_direct.parallel_prepare()


class ConvolutionDensityMatricesFromDisk(ConvolutionDensityMatrices):

    """
    Collection of density matrices in the Kohn-Sham basis after convolution with
    one or several pulses, for different times. Read from disk.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    pulserho_fmt
        Formatting string for the density matrices saved to disk.

        The formatting string should be a plain string containing variable
        placeholders within curly brackets `{}`. It should not be confused with
        a formatted string literal (f-string).

        Example:

        * pulserho_fmt =  `pulserho/t{time:09.1f}{tag}.npy`.

        Accepts variables

        * `{time}` - Time in units of as.
        * `{tag}` - Derivative tag, `''`, `'-Iomega'`, or `'-omega2'`.
        * `{pulsefreq}` - Pulse frequency in units of eV.
        * `{pulsefwhm}` - Pulse FWHM in units of fs.
    pulses
        Convolute the density matrices with these pulses.
    times
        Compute density matrices for these times (or as close to them as possible). In units of as.
    derivative_order_s
        Compute density matrix derivatives of the following orders.
        `0` for plain density matrix and positive integers for derivatives.
    real
        Calculate real part of density matrices.
    imag
        Calculate imaginary part of density matrices.
    calc_size
        Size of the calculation communicator.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 pulserho_fmt: str,
                 pulses: Collection[PerturbationLike],
                 times: list[float] | Array1D[np.float64],
                 derivative_order_s: list[int] = [0],
                 log: Logger | None = None,
                 real: bool = True,
                 imag: bool = True,
                 calc_size: int = 1):
        reader = TimeDensityMatrixReader(pulserho_fmt=pulserho_fmt,
                                         ksd=ksd,
                                         pulses=pulses,
                                         filter_times=times,
                                         derivative_order_s=derivative_order_s,
                                         log=log)
        self.reader = reader
        super().__init__(ksd=reader.ksd, pulses=reader.pulses, times=reader.times,
                         real=real, imag=imag,
                         derivative_order_s=derivative_order_s, calc_size=calc_size, log=reader.log)
        self.ksd.distribute(self.calc_comm)
        self.pulserho_fmt = pulserho_fmt

    def __str__(self) -> str:
        return str(self.reader)

    def __iter__(self) -> Generator[tuple[ConvolutionDensityMatrixMetadata, DensityMatrix], None, None]:
        for work in self.local_work_plan:
            self.log.start('read')
            matrices: dict[int, NDArray[np.complex128] | None] = dict()
            for derivative in self.derivative_order_s:
                if self.calc_comm.rank > 0:
                    # Don't read on non calc comm root ranks
                    matrices[derivative] = None
                    continue

                matrices[derivative] = self.reader.read(work.time, work.pulse, derivative)

            dm = DensityMatrix(ksd=self.ksd, matrices=matrices, comm=self.calc_comm)

            yield work, dm


class ConvolutionDensityMatricesFromFrequency(ConvolutionDensityMatrices):

    """
    Collection of density matrices in the Kohn-Sham basis after convolution with
    one or several pulses, for different times. Obtained from the the density
    matrices in frequency space, which are read from disk.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    frho_fmt
        Formatting string for the density matrices
        in frequency space saved to disk.

        The formatting string should be a plain string containing variable
        placeholders within curly brackets `{}`. It should not be confused with
        a formatted string literal (f-string).

        Example:

        * frho_fmt = `frho/w{freq:05.2f}-{reim}.npy'`.

        Accepts variables:

        * `{freq}` - Frequency in units of eV.
        * `{reim}` - `'Re'` or `'Im'` for Fourier transform of real/imaginary
          part of density matrix.
    perturbation
        Perturbation that was present during time propagation.
    pulses
        Convolute the density matrices with these pulses.
    times
        Compute density matrices for these times (or as close to them as possible). In units of as.
    derivative_order_s
        Compute density matrix derivatives of the following orders.
        `0` for plain density matrix and positive integers for derivatives.
    real
        Calculate real part of density matrices.
    imag
        Calculate imaginary part of density matrices.
    calc_size
        Size of the calculation communicator.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 frho_fmt: str,
                 perturbation: PerturbationLike,
                 pulses: Collection[PerturbationLike],
                 times: list[float] | NDArray[np.float64],
                 derivative_order_s: list[int] = [0],
                 log: Logger | None = None,
                 real: bool = True,
                 imag: bool = True,
                 calc_size: int = 1):
        super().__init__(ksd=ksd, pulses=pulses, times=times,
                         real=real, imag=imag,
                         derivative_order_s=derivative_order_s, calc_size=calc_size, log=log)
        # Frequency grid for convolution
        if not all(pulse.todict()['name'] == 'GaussianPulse' for pulse in self.pulses):
            raise NotImplementedError('Only Gaussian pulses implemented for ResponseFromFourierTransform')

        freq_spacing = 0.05  # 50meV
        freq_min = min((pulse.pulse.omega0 - 4 * pulse.pulse.sigma) * au_to_eV  # type: ignore
                       for pulse in self.pulses)
        freq_min = max(freq_min, freq_spacing)
        freq_max = min((pulse.pulse.omega0 + 4 * pulse.pulse.sigma) * au_to_eV  # type: ignore
                       for pulse in self.pulses)

        transformer = ExactFourierTransformer.from_file(
                frho_fmt=frho_fmt, ksd=ksd, perturbation=perturbation,
                real=real, imag=imag, log=self.log, comm=self.calc_comm,
                freq_spacing=freq_spacing, freq_min=freq_min, freq_max=freq_max)
        self.transformer = transformer

    def __str__(self) -> str:
        lines = ['Response from Fourier transform of density matrices on disk']
        lines.append('')
        lines.append(f'Calculating response for {self.nt} times and {len(self.pulses)} pulses')
        lines.append(f'  times: {format_times(self.times)}')

        return '\n'.join(lines)

    def __iter__(self) -> Generator[tuple[ConvolutionDensityMatrixMetadata, DensityMatrix], None, None]:
        self.transformer.parallel_prepare()

        for work in self.local_work_plan:
            matrices: dict[int, NDArray[np.complex128] | None] = dict()

            for derivative in self.derivative_order_s:
                buffer = self.transformer.convolve(work.time, work.pulse, derivative)
                if self.calc_comm.rank > 0:
                    matrices[derivative] = None
                    continue
                assert buffer is not None
                # Buffer shape is i, a, pulses, times
                if 'Re' in self.reim:
                    Rerho_ia = buffer._get_real(derivative)
                if 'Im' in self.reim:
                    Imrho_ia = buffer._get_imag(derivative)
                if 'Re' in self.reim and 'Im' in self.reim:
                    # Complex result
                    rho_ia = Rerho_ia + 1j * Imrho_ia
                elif 'Re' in self.reim:
                    # Real result
                    rho_ia = Rerho_ia
                else:
                    rho_ia = 1j * Imrho_ia
                matrices[derivative] = rho_ia

            dm = DensityMatrix(ksd=self.ksd, matrices=matrices, comm=self.calc_comm)

            yield work, dm


class ConvolutionDensityMatricesFromWaveFunctions(ConvolutionDensityMatrices):

    """
    Collection of density matrices in the Kohn-Sham basis after convolution with
    one or several pulses, for different times. Obtained from the time-dependent wave functions file,
    which is read from disk.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    wfs_fname
        File name of the time-dependent wave functions file.
    perturbation
        Perturbation that was present during time propagation.
    pulses
        Convolute the density matrices with these pulses.
    times
        Compute density matrices for these times (or as close to them as possible). In units of as.
    derivative_order_s
        Compute density matrix derivatives of the following orders.
        `0` for plain density matrix and positive integers for derivatives.
    real
        Calculate real part of density matrices.
    imag
        Calculate imaginary part of density matrices.
    calc_size
        Size of the calculation communicator.
    stridet
        Skip this many steps when reading the time-dependent wave functions file.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 wfs_fname: str,
                 perturbation: PerturbationLike,
                 pulses: Collection[PerturbationLike],
                 times: list[float] | Array1D[np.float64],
                 derivative_order_s: list[int] = [0],
                 real: bool = True,
                 imag: bool = True,
                 calc_size: int = 1,
                 log: Logger | None = None,
                 stridet: int = 1):
        _, calc_size = two_communicator_sizes(-1, calc_size)
        # The calc_comm rank 0's are world ranks 0, with a spacing of calc_size
        result_on_ranks = list(range(0, world.size, calc_size))

        rho_nn_conv = PulseConvolver.from_parameters(
                wfs_fname, ksd,
                pulses=pulses,
                perturbation=perturbation,
                yield_re=real, yield_im=imag,
                filter_times=np.array(times) * as_to_au,
                derivative_order_s=list(sorted(derivative_order_s)),
                stridet=stridet,
                result_on_ranks=result_on_ranks,
                log=log)
        self.rho_nn_conv = rho_nn_conv

        super().__init__(ksd=rho_nn_conv.ksd, pulses=pulses, times=rho_nn_conv.time_t * au_to_as,
                         real=real, imag=imag,
                         derivative_order_s=derivative_order_s, calc_size=calc_size,
                         log=rho_nn_conv.log)

    def __str__(self) -> str:
        lines = []
        lines.append('Response from time-dependent wave functions ')
        lines.append('')
        lines += ['' + line for line in str(self.rho_nn_conv.rho_nn_reader.rho_wfs_reader).split('\n')]
        lines.append('')
        lines += ['' + line for line in str(self.rho_nn_conv.rho_nn_reader).split('\n')]
        lines.append('')
        lines += ['' + line for line in str(self.rho_nn_conv).split('\n')]

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        return self.rho_nn_conv.get_memory_estimate()

    @property
    def myt(self) -> list[int]:
        """ List of indices corresponding to the time indices on held on this rank. """
        return self.rho_nn_conv.my_work()

    def work_loop(self,
                  rank: int) -> Generator[ConvolutionDensityMatrixMetadata | None, None, None]:
        nt = len(self.times)
        ntperrank = (nt + self.loop_comm.size - 1) // self.loop_comm.size

        for p, pulse in enumerate(self.pulses):
            for localt in range(ntperrank):
                globalt = rank + localt * self.loop_comm.size
                if globalt < nt:
                    yield ConvolutionDensityMatrixMetadata(density_matrices=self, globalt=globalt,
                                                           localt=localt, globalp=p, localp=p)
                else:
                    yield None

    def __iter__(self) -> Generator[tuple[ConvolutionDensityMatrixMetadata, DensityMatrix], None, None]:
        parameters = self.rho_nn_conv.rho_nn_reader._parameters
        flt = (slice(parameters.n1size), slice(parameters.n2size))

        dist_buffer = self.rho_nn_conv.dist_buffer  # Perform the redistribution
        self.ksd.distribute(self.calc_comm)

        if self.calc_comm.rank != 0:
            assert len(self.myt) == 0, self.myt

        for work in self.local_work_plan:
            if self.calc_comm.rank == 0:
                assert self.myt[work.localt] == work.globalt
            localflt = flt + (work.localp, work.localt)

            matrices: dict[int, NDArray[np.complex128] | None] = dict()
            for derivative in self.derivative_order_s:
                if self.calc_comm.rank > 0:
                    matrices[derivative] = None
                    continue
                # Buffer shape is i, a, pulses, times
                if 'Re' in self.reim:
                    Rerho_ia = dist_buffer._get_real(derivative)[localflt]
                if 'Im' in self.reim:
                    Imrho_ia = dist_buffer._get_imag(derivative)[localflt]
                if 'Re' in self.reim and 'Im' in self.reim:
                    # Complex result
                    # Compared to numpy, we use another convention, hence the minus sign
                    rho_ia = Rerho_ia - 1j * Imrho_ia
                elif 'Re' in self.reim:
                    # Real result
                    rho_ia = Rerho_ia
                else:
                    rho_ia = -1j * Imrho_ia
                matrices[derivative] = rho_ia
            dm = DensityMatrix(ksd=self.ksd, matrices=matrices, comm=self.calc_comm)

            yield work, dm

    def parallel_prepare(self):
        self.rho_nn_conv.dist_buffer  # Perform the redistribution


class ExactFourierTransformer:

    def __init__(self,
                 reader: FrequencyDensityMatrixReader,
                 perturbation: PerturbationLike,
                 real: bool = True,
                 imag: bool = True,
                 comm: Communicator | None = None,
                 log: Logger | None = None):
        if comm is None:
            comm = world
        self._comm = comm
        self.reader = reader
        if log is None:
            log = Logger()
        self.log = log

        self.real = real
        self.imag = imag

        self.frequency_buffer = DensityMatrixBuffer(self.nnshape, (self.reader.nw, ), np.complex128)
        self._ready = False

        omega_w = self.reader.frequencies * eV_to_au

        self.perturbation = create_perturbation(perturbation)
        assert isinstance(self.perturbation, DeltaKick)

        # Need a uniform spacing for Simpson's integration rule
        dw = omega_w[1] - omega_w[0]
        nw = len(omega_w)
        if not np.allclose(omega_w[1:] - dw, omega_w[:-1]):
            raise ValueError('Variable frequency spacing.')
        # integration weights from Simpson's integration rule
        self._weight_w = dw / 3 * np.array([1] + [4, 2] * int((nw - 2) / 2)
                                           + [4] * (nw % 2) + [1])

        self._weight_w *= 2 / (2 * np.pi)
        self._weight_w /= self.perturbation.strength

    @property
    def comm(self) -> Communicator:
        """ MPI commonicator.

        During convolution, the density matrices are distributed with
        different chunks on different ranks of this communicator.
        """
        return self._comm

    @property
    def full_nnshape(self) -> tuple[int, int]:
        imin, imax, amin, amax = [int(i) for i in self.reader.ksd.ialims()]
        return (imax - imin + 1, amax - amin + 1)

    @property
    def nnshape(self) -> tuple[int, int]:
        stridei = (self.full_nnshape[0] + self.comm.size - 1) // self.comm.size
        return (stridei, self.full_nnshape[1])

    @property
    def myslice(self) -> slice:
        return self.slice_of_rank(self.comm.rank)

    def slice_of_rank(self, rank: int) -> slice:
        stridei = (self.full_nnshape[0] + self.comm.size - 1) // self.comm.size
        return slice(rank * stridei, (rank + 1) * stridei)

    def parallel_prepare(self):
        # Read all the Fourier transform of density matrices
        self.frequency_buffer.zero_buffers(real=self.real, imag=self.imag, derivative_order_s=[0])
        for w, freq in enumerate(self.reader.frequencies):
            if self.real:
                rho_ia = self.reader.read(frequency=freq, real=True)[self.myslice]
                self.frequency_buffer[w].real[:len(rho_ia)] = rho_ia
            if self.imag:
                rho_ia = self.reader.read(frequency=freq, real=False)[self.myslice]
                self.frequency_buffer[w].imag[:len(rho_ia)] = rho_ia
        self._ready = True

    def convolve(self,
                 time: float,
                 pulse: Perturbation,
                 derivative: int) -> DensityMatrixBuffer | None:
        if not self._ready:
            self.parallel_prepare()

        assert isinstance(pulse, PulsePerturbation)
        omega_w = self.reader.frequencies * eV_to_au
        pulse_w = pulse.pulse.fourier(omega_w)

        buffer = DensityMatrixBuffer(self.frequency_buffer.nnshape, (), dtype=np.float64)

        self.log.start('convolve')
        exp_w = np.exp(-1j * omega_w * time * as_to_au)
        exp_w *= self._weight_w * pulse_w * (-1j * omega_w) ** derivative

        if self.real:
            # Real part
            rho_iaw = self.frequency_buffer.real[:]
            conv_rho_ia = rho_iaw.real @ exp_w.real
            conv_rho_ia -= rho_iaw.imag @ exp_w.imag

            buffer.store(True, derivative, conv_rho_ia)

        if self.imag:
            # Imaginary part
            rho_iaw = self.frequency_buffer.imag[:]
            conv_rho_ia = rho_iaw.real @ exp_w.real
            conv_rho_ia -= rho_iaw.imag @ exp_w.imag

            buffer.store(False, derivative, conv_rho_ia)

        if self.comm.rank == 0:
            full_buffer = DensityMatrixBuffer(self.frequency_buffer.nnshape, (), dtype=np.float64)
            full_buffer.zero_buffers(real=self.real, imag=self.imag, derivative_order_s=[derivative])

        # Send to root
        for rank in range(self.comm.size):
            if rank == self.comm.rank:
                # Send own work to root
                buffer.send_arrays(self.comm, 0)

            if self.comm.rank == 0:
                # Receive on root, fill buffer
                buffer.recv_arrays(self.comm, rank)
                for full_rho_ia, part_rho_ia in zip(full_buffer._iter_buffers(),
                                                    buffer._iter_buffers()):
                    full_slice_ia = full_rho_ia[self.slice_of_rank(rank)]
                    full_slice_ia[:] = part_rho_ia[:len(full_slice_ia)]

        if self.comm.rank == 0:
            self.log(f'Convolved density matrices in {self.log.elapsed("convolve"):.1f} s',
                     who='Response', flush=True)

        if self.comm.rank == 0:
            return full_buffer

        return None

    @classmethod
    def from_file(cls,
                  frho_fmt: str,
                  ksd: KohnShamDecomposition | str,
                  perturbation: PerturbationLike,
                  freq_min: float,
                  freq_spacing: float,
                  freq_max: float,
                  real: bool = True,
                  imag: bool = True,
                  comm: Communicator | None = None,
                  log: Logger | None = None) -> ExactFourierTransformer:
        frequencies = np.arange(freq_min, freq_max + freq_spacing * 0.1, freq_spacing)

        reader = FrequencyDensityMatrixReader(frho_fmt=frho_fmt,
                                              ksd=ksd,
                                              filter_frequencies=frequencies,
                                              real=real, imag=imag,
                                              comm=None,  # Look for files on world
                                              log=log)
        if reader.frequencies[0] > freq_min:
            raise ValueError(f'Could not satisfy lower bound for frequency grid with files matching {frho_fmt}. '
                             f'Found {reader.frequencies[0]:.2f}eV, reqested {freq_min:.2f}eV.')

        if reader.frequencies[-1] > freq_max:
            raise ValueError(f'Could not satisfy upper bound for frequency grid with files matching {frho_fmt}. '
                             f'Found {reader.frequencies[-1]:.2f}eV, reqested {freq_max:.2f}eV.')

        dw = reader.frequencies[1] - reader.frequencies[0]
        if dw > freq_spacing + 1e-8:
            raise ValueError(f'Could not construct sufficiently dense frequency grid with files matching {frho_fmt}. '
                             f'Found spacing of {dw:.4f}eV, reqested {freq_spacing:.4f}eV.')

        return cls(reader=reader, perturbation=perturbation, real=real, imag=imag, comm=comm, log=log)



from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Generator, Collection

import numpy as np

from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition

from ..perturbation import create_perturbation, Perturbation, PerturbationLike
from ..density_matrices.density_matrix import DensityMatrix
from ..density_matrices.frequency import (FrequencyDensityMatrixMetadata,
                                          FrequencyDensityMatrices)
from ..density_matrices.time import (ConvolutionDensityMatrixMetadata,
                                     ConvolutionDensityMatrices)
from ..typing import Array1D
from ..utils import add_fake_kpts, Logger


class BaseResponse(ABC):

    """ Object describing response; obtained from :term:`TDDFT` calculation.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    perturbation
        The perturbation that was present during the TDDFT calculation.
        None to mark it as an ne perturbation.
    calc_size
        Size of the calculation communicator.
    """
    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 perturbation: PerturbationLike = None,
                 calc_size: int = 1):
        if isinstance(ksd, KohnShamDecomposition):
            self._ksd = ksd
        else:
            self._ksd = KohnShamDecomposition(filename=ksd)
        add_fake_kpts(self._ksd)

        self._perturbation = create_perturbation(perturbation)
        self.calc_size = calc_size

    def __str__(self) -> str:
        lines = [f'{self.__class__.__name__}']
        lines += [f'  ksd: {self.ksd.filename if self.ksd.filename is not None else "From calc"}']
        lines += ['  perturbation:']
        lines += ['    ' + line for line in str(self.perturbation).split('\n')]
        return '\n'.join(lines)

    @property
    def ksd(self) -> KohnShamDecomposition:
        """ Kohn-Sham decomposition object. """
        return self._ksd

    @property
    def perturbation(self) -> Perturbation:
        """ The perturbation that caused this response. """
        return self._perturbation

    @abstractmethod
    def _get_time_density_matrices(self,
                                   times: list[float] | Array1D[np.float64],
                                   pulses: Collection[PerturbationLike],
                                   derivative_order_s: list[int] = [0],
                                   real: bool = True,
                                   imag: bool = True,
                                   log: Logger | None = None,
                                   ) -> ConvolutionDensityMatrices:
        raise NotImplementedError

    @abstractmethod
    def _get_frequency_density_matrices(self,
                                        frequencies: list[float] | Array1D[np.float64],
                                        frequency_broadening: float = 0,
                                        real: bool = True,
                                        imag: bool = True,
                                        log: Logger | None = None,
                                        ) -> FrequencyDensityMatrices:
        raise NotImplementedError

    def iterate_density_matrices_in_time(self,
                                         times: list[float] | Array1D[np.float64],
                                         pulses: Collection[PerturbationLike],
                                         derivative_order_s: list[int] = [0],
                                         real: bool = True,
                                         imag: bool = True,
                                         log: Logger | None = None,
                                         ) -> Generator[tuple[ConvolutionDensityMatrixMetadata,
                                                              DensityMatrix], None, None]:
        """ Obtain density matrices at the given times in response to the given pulses.

        If the given pulse(s) differ from the perturbation that caused this response,
        then the pulse convolution trick is applied to obtain the response to the given
        pulse(s).

        Parameters
        ----------
        times
            Calculate density matrices for these times (or as close to them as possible). In units of as.
        pulses
            Calculate density matrices in response to these pulses.
        derivative_order_s
            Calculate density matrix derivatives of the following orders.
            ``0`` for plain density matrix and positive integers for derivatives.
        real
            Calculate the real part of density matrices.
        imag
            Calculate the imaginary part of density matrices.
        log
            Logger object.

        Yields
        ------
        Tuple (work, dm) on the root rank of the calculation communicator:

            work
                An object representing the metadata (time and pulse) for the work done.
            dm
                Density matrix for this time and pulse.
        """
        density_matrices = self._get_time_density_matrices(
                times, pulses, derivative_order_s, real, imag, log)
        yield from density_matrices

    def iterate_density_matrices_in_frequency(self,
                                              frequencies: list[float] | Array1D[np.float64],
                                              frequency_broadening: float = 0,
                                              real: bool = True,
                                              imag: bool = True,
                                              log: Logger | None = None,
                                              ) -> Generator[tuple[FrequencyDensityMatrixMetadata,
                                                                   DensityMatrix], None, None]:
        """ Obtain density matrices at the given frequencies.

        Parameters
        ----------
        frequencies
            Compute density matrices for these frequencies (or as close to them as possible). In units of eV.
        frequency_broadening
            Gaussian broadening width in atomic units. Default (0) is no broadening.
        real
            Calculate the Fourier transform of the real part of the density matrix.
        imag
            Calculate the Fourier transform of the imaginary part of the density matrix.

        Yields
        ------
        Tuple (work, dm) on the root rank of the calculation communicator:

            work
                An object representing the metadata (frequency) for the work done.
            dm
                Density matrix for this frequency.
        """
        density_matrices = self._get_frequency_density_matrices(
                frequencies, frequency_broadening, real, imag, log)
        yield from density_matrices

    def write_in_time(self,
                      pulserho_fmt: str,
                      times: list[float] | Array1D[np.float64],
                      pulses: Collection[PerturbationLike],
                      derivative_order_s: list[int] = [0],
                      real: bool = True,
                      imag: bool = True,
                      log: Logger | None = None):
        density_matrices = self._get_time_density_matrices(
                times, pulses, derivative_order_s, real, imag, log)
        density_matrices.write_to_disk(pulserho_fmt)

    def write_in_frequency(self,
                           frho_fmt: str,
                           frequencies: list[float] | Array1D[np.float64],
                           frequency_broadening: float = 0,
                           real: bool = True,
                           imag: bool = True,
                           log: Logger | None = None):
        density_matrices = self._get_frequency_density_matrices(
                frequencies, frequency_broadening, real, imag, log)
        density_matrices.write_to_disk(frho_fmt)



from __future__ import annotations

from .base import BaseResponse
from .gpaw import ResponseFromWaveFunctions
from .numpy import ResponseFromDensityMatrices, ResponseFromFourierTransform

__all__ = [
    'BaseResponse',
    'ResponseFromWaveFunctions',
    'ResponseFromFourierTransform',
    'ResponseFromDensityMatrices',
]



from __future__ import annotations


import numpy as np
from numpy.typing import NDArray

from .. import __version__
from ..perturbation import create_perturbation, PerturbationLike


def write_spectrum(out_fname: str,
                   frequencies: list[float] | NDArray[np.float64],
                   spectrum: NDArray[np.float64],
                   frequency_broadening: float,
                   total_time: float,
                   timestep: float,
                   perturbation: PerturbationLike):
    """ Write dipole strength function (spectrum) to text file.

    Parameters
    ----------
    out_fname
       File name of the written file.
    frequencies
        Array of frequencies in units of eV.
    spectrum
        Spectrum corresponding to the :attr:`frequencies` in units of 1/eV.
    frequency_broadening
        Gaussian broadening width in units of eV. Default (0) is no broadening.
    total_time
        Total simulation time in units of as.
    timestep
        Timestep in units of as.
    perturbation
        The perturbation that was applied in the :term:`TDDFT` calculation.
    """
    frequencies = np.array(frequencies)
    osc_wv = np.array(spectrum)
    perturbation = create_perturbation(perturbation)

    if frequency_broadening == 0:
        broadening = 'No broadening'
    else:
        broadening = f'Gaussian broadening width {frequency_broadening:.2f} eV'
    perturbationstr = '\n'.join(['Perturbation during calculation was:'] +
                                ['  ' + line for line in str(perturbation).split('\n')])

    header = ('Photoabsorption spectrum from real-time propagation\n'
              f'Calculated using rhodent version: {__version__}\n'
              f'Total time = {total_time*1e-3:.4f} fs, Time steps = {timestep:.2f} as\n'
              f'{perturbationstr}\n'
              f'{broadening}\n'
              f'{"om (eV)":>10} {"S_x (1/eV)":>20} {"S_y (1/eV)":>20} {"S_z (1/eV)":>20}'
              )
    data_wi = np.zeros((len(frequencies), 1 + 3))
    data_wi[:, 0] = frequencies
    data_wi[:, 1:] = osc_wv
    fmt = '%12.6f' + (' %20.10le' * 3)
    np.savetxt(str(out_fname), data_wi, header=header, fmt=fmt)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Callable, Generator

import numpy as np
from numpy.typing import NDArray
from itertools import zip_longest

from ase.units import Bohr
from ase.io.ulm import Reader

from gpaw.mpi import broadcast, world
from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition
from gpaw.lcaotddft.wfwriter import WaveFunctionReader

from ..buffer import DensityMatrixBuffer
from ...utils import Logger, add_fake_kpts, get_array_filter
from ...utils.logging import format_times
from ...utils.memory import HasMemoryEstimate, MemoryEstimate
from ...typing import Array1D, Communicator


class BaseWfsReader(ABC):

    """ Read wave functions or density matrices from the time-dependent wave functions file.

    Parameters
    ----------
    wfs_fname
        File name of the time-dependent wave functions file written by GPAW.
    comm
        MPI communicator.
    yield_re
        Whether to read and yield the real part of wave functions/density matrices.
    yield_im
        Whether to read and yield the imaginary part of wave functions/density matrices.
    stridet
        Skip this many steps when reading.
    tmax
        Last time index to read.
    filter_times
        A list of times to read in atomic units. The closest times in the time-dependent
        wave functions file will be read. Applied after skipping the stridet steps to tmax.
    log
        Logger object.
    """

    def __init__(self,
                 wfs_fname: str,
                 comm=world,
                 yield_re: bool = True,
                 yield_im: bool = True,
                 stridet: int = 1,
                 tmax: int = -1,
                 filter_times: list[float] | Array1D[np.float64] | None = None,
                 log: Logger | None = None):
        self._comm = comm
        self._yield_re = yield_re
        self._yield_im = yield_im
        if log is None:
            log = Logger()
        self._log = log

        # The main reader is closed when it is garbage collected
        # Hence, we need to keep it in the scope
        self.mainreader = WaveFunctionReader(wfs_fname)
        self._time_t, self.initreader, self._full_reader_t = prepare_wave_function_readers(
            self.mainreader, comm, self.log, stridet=stridet, tmax=tmax)
        self._flt_t = get_array_filter(self._time_t, filter_times)
        self.reader_t = [self._full_reader_t[r]
                         for r in np.arange(len(self._full_reader_t), dtype=int)[self._flt_t]]

    @property
    def comm(self) -> Communicator:
        """ MPI communicator. """
        return self._comm

    @property
    def yield_re(self) -> bool:
        """ Whether this object should read real parts. """
        return self._yield_re

    @property
    def yield_im(self) -> bool:
        """ Whether this object should read imaginary parts. """
        return self._yield_im

    @property
    def log(self) -> Logger:
        """ Logger object. """
        return self._log

    @property
    def time_t(self) -> Array1D[np.float64]:
        """ Array of times to read; in atomic units. """
        return self._time_t[self._flt_t]  # type: ignore

    @property
    def nt(self) -> int:
        """ Number of times to read. """
        return len(self.time_t)

    @property
    def dt(self) -> float:
        """ Time step in atomic units. """
        time_t = self.time_t
        dt = time_t[1] - time_t[0]
        if not np.allclose(time_t[1:] - dt, time_t[:-1]):
            fname = self.mainreader.filename
            raise ValueError(f'Unable to get a time step. Variable time step in {fname}.')

        return dt

    def work_loop(self,
                  rank: int) -> Generator[int | None, None, None]:
        """ Yield the time indices that this rank will read.

        New indices are yielded until the end of self.reader_t is reached
        (across all ranks).

        Yields
        ------
        Time index between 0 and len(self.reader_t) - 1 corresponding to
        the time being read by this rank. Or None if this rank has nothing
        to read.
        """
        for t_r in self.work_loop_by_ranks():
            yield t_r[rank]

    def work_loop_by_ranks(self) -> Generator[list[int], None, None]:
        nt = self.nt
        ntperrank = (nt + self.comm.size - 1) // self.comm.size

        for localt in range(ntperrank):
            globalt_r = [rank + localt * self.comm.size for rank in range(self.comm.size)]
            globalt_r = [globalt if globalt < nt else None for globalt in globalt_r]
            yield globalt_r

    def global_work_loop(self) -> Generator[int, None, None]:
        for chunks_r in self.work_loop_by_ranks():
            for chunk in chunks_r:
                if chunk is None:
                    continue
                yield chunk

    @abstractmethod
    def iread(self, *args, **kwargs) -> Generator[DensityMatrixBuffer, None, None]:
        """ Iteratively read wave functions or density matrices time by time. """
        raise NotImplementedError

    @abstractmethod
    def nnshape(self, *args, **kwargs) -> tuple[int, int]:
        """ Shape of the density matrices or wave functions. """
        raise NotImplementedError

    def gather_on_root(self, *args, **kwargs) -> Generator[DensityMatrixBuffer | None, None, None]:
        for indices_r, dm_buffer in zip_longest(self.work_loop_by_ranks(),
                                                self.iread(*args, **kwargs), fillvalue=None):
            assert indices_r is not None, 'Work loop shorter than work'

            # Yield root's own work
            if self.comm.rank == 0:
                assert indices_r[0] is not None
                assert dm_buffer is not None
                dm_buffer.ensure_contiguous_buffers()

                yield dm_buffer.copy()
            else:
                yield None

            # Yield the work of non-root
            for recvrank, recvindices in enumerate(indices_r[1:], start=1):
                if recvindices is None:
                    # No work on this recvrank
                    continue

                if self.comm.rank == 0:
                    # Receive work
                    assert dm_buffer is not None
                    dm_buffer.recv_arrays(self.comm, recvrank, log=self.log)
                    yield dm_buffer.copy()
                else:
                    # Send work to root if there is any
                    if self.comm.rank == recvrank:
                        assert dm_buffer is not None
                        dm_buffer.send_arrays(self.comm, 0, log=self.log)
                    yield None

    def collect_on_root(self, *args, **kwargs) -> DensityMatrixBuffer | None:
        nnshape = self.nnshape(*args, **kwargs)
        full_dm = DensityMatrixBuffer(nnshape, (self.nt, ), np.float64)
        if self.yield_re:
            full_dm.zeros(True, 0)
        if self.yield_im:
            full_dm.zeros(False, 0)

        for t, dm_buffer in zip_longest(self.global_work_loop(),
                                        self.gather_on_root(*args, **kwargs), fillvalue=None):
            if self.comm.rank != 0:
                continue

            assert t is not None, 'Iterators must be same length'
            assert dm_buffer is not None, 'Iterators must be same length'

            for partial_data_nn, full_data_nn in zip(dm_buffer._iter_buffers(),
                                                     full_dm[t]._iter_buffers()):
                full_data_nn[:] += partial_data_nn

        if self.comm.rank != 0:
            return None

        return full_dm


class KohnShamRhoWfsReader(HasMemoryEstimate, BaseWfsReader):

    """ Read density matrices from the time-dependent wave functions file.

    Yield density matrices time by time.

    Parameters
    ----------
    wfs_fname
        File name of the time-dependent wave functions file.
    ksd
        KohnShamDecomposition object or file name to the ksd file.
    comm
        MPI communicator.
    yield_re
        Whether to read and yield the real part of wave functions/density matrices.
    yield_im
        Whether to read and yield the imaginary part of wave functions/density matrices.
    stridet
        Skip this many steps when reading.
    tmax
        Last time index to read.
    filter_times
        A list of times to read in atomic units. The closest times in the time-dependent wave functions file
        will be read.
    striden
        Option passed through to the LCAORhoWfsReader.
    log
        Logger object.
    """

    def __init__(self,
                 wfs_fname: str,
                 ksd: str | KohnShamDecomposition,
                 comm=world,
                 yield_re: bool = True,
                 yield_im: bool = True,
                 stridet: int = 1,
                 tmax: int = -1,
                 filter_times: list[float] | Array1D[np.float64] | None = None,
                 log: Logger | None = None,
                 striden: int = 0):
        # Set up an internal LCAO density matrix reader
        self.lcao_rho_reader = LCAORhoWfsReader(
            wfs_fname=wfs_fname, comm=comm,
            yield_re=yield_re, yield_im=yield_im, log=log,
            stridet=stridet, tmax=tmax,
            filter_times=filter_times, striden=striden)

        # And copy its attributes to self
        self._yield_re = yield_re
        self._yield_im = yield_im
        self._comm = self.lcao_rho_reader.comm
        self._log = self.lcao_rho_reader.log
        self.mainreader = self.lcao_rho_reader.mainreader
        self.initreader = self.lcao_rho_reader.initreader
        self._full_reader_t = self.lcao_rho_reader._full_reader_t
        self.reader_t = self.lcao_rho_reader.reader_t
        self._flt_t = self.lcao_rho_reader._flt_t
        self._time_t = self.lcao_rho_reader._time_t

        # Set up ksd
        if isinstance(ksd, KohnShamDecomposition):
            self._ksd = ksd
        else:
            self._ksd = KohnShamDecomposition(filename=ksd)
            add_fake_kpts(self._ksd)

        self._C0S_sknM: NDArray[np.float64] | None = None
        self._rho0_sknn: NDArray[np.float64] | None = None

    def __str__(self) -> str:
        nn, nM = proxy_coefficients(self.initreader).shape[2:]
        ntperrank = (self.nt + self.comm.size - 1) // self.comm.size

        lines = []
        lines.append('Time-dependent wave functions reader')
        lines.append('  Constructing density matrices in basis of ground state orbitals.')
        lines.append('')
        lines.append(f'  file: {self.mainreader.filename}')
        lines.append(f'    wave function dimensions {(nn, nM)}')
        lines.append(f'    {self.nt} times')
        lines.append(f'      {format_times(self.time_t, units="au")}')
        lines.append(f'  {self.comm.size} ranks reading in {ntperrank} iterations')

        return '\n'.join(lines)

    def get_memory_estimate(self) -> MemoryEstimate:
        nn, nM = proxy_coefficients(self.initreader).shape[2:]

        memory_estimate = MemoryEstimate()
        memory_estimate.add_key('C0S_nM', (nn, nM), float,
                                on_num_ranks=self.comm.size)
        memory_estimate.add_key('rho0_MM', (nM, nM), float,
                                on_num_ranks=self.comm.size)

        return memory_estimate

    @property
    def ksd(self) -> KohnShamDecomposition:
        """ Kohn-Sham decomposition object. """
        return self._ksd

    @property
    def C0S_sknM(self) -> NDArray[np.float64]:
        if self._C0S_sknM is None:
            self.log(f'Constructing C0_sknM on {self.comm.size} ranks',
                     who='Reader', rank=0, flush=True)
            self._C0S_sknM = read_C0S_parallel(self.ksd.reader, comm=self.comm)
            self.log('Constructed C0_sknM',
                     who='Reader', rank=0, flush=True)
            assert self._C0S_sknM is not None
        return self._C0S_sknM

    @property
    def rho0_sknn(self) -> NDArray[np.float64]:
        if self._rho0_sknn is None:
            f_skn = self.ksd.reader.occ_un[:]
            nn = f_skn.shape[2]
            rho0_sknn = np.zeros(f_skn.shape[:2] + (nn, nn))
            diag_nn = np.eye(nn, dtype=bool)
            rho0_sknn[..., diag_nn] = f_skn
            self._rho0_sknn = rho0_sknn
        return self._rho0_sknn

    def nnshape(self,
                s: int,
                k: int,
                n1: slice,
                n2: slice) -> tuple[int, int]:
        n1size = n1.stop - n1.start
        n2size = n2.stop - n2.start
        nnshape = (n1size, n2size)
        return nnshape

    def parallel_prepare(self):
        """ Read everything necessary synchronously on all ranks. """
        self.C0S_sknM
        self.lcao_rho_reader.rho0_skMM

    def iread(self,
              s: int,
              k: int,
              n1: slice,
              n2: slice) -> Generator[DensityMatrixBuffer, None, None]:
        """ Read the density matrices time by time.

        Parameters
        ----------
        s, k, n1, n2
            Read these indices.
        """
        dm_buffer = DensityMatrixBuffer(self.nnshape(s, k, n1, n2), (), np.float64)
        if self.yield_re:
            dm_buffer.zeros(True, 0)
        if self.yield_im:
            dm_buffer.zeros(False, 0)

        einsumstr = 'nN,mM,NM->nm'
        self.C0S_sknM  # Read this on all ranks

        nM = self.C0S_sknM.shape[3]
        sliceM = slice(0, nM)

        for lcao_dm in self.lcao_rho_reader.iread(s, k, sliceM, sliceM):
            C0S_nM1 = self.C0S_sknM[s, k, n1, :]  # Here n is full KS basis
            C0S_nM2 = self.C0S_sknM[s, k, n2, :]

            self.log.start('read')

            # Conjugate C_nM2
            if self.yield_re:
                Rerho_MM = lcao_dm.real
                Rerho_x = np.einsum(einsumstr, C0S_nM1, C0S_nM2, Rerho_MM, optimize=True)
                dm_buffer.safe_fill(True, 0, Rerho_x)
            if self.yield_im:
                Imrho_MM = lcao_dm.imag
                Imrho_x = np.einsum(einsumstr, C0S_nM1, C0S_nM2, Imrho_MM, optimize=True)
                dm_buffer.safe_fill(False, 0, Imrho_x)

            yield dm_buffer


class LCAORhoWfsReader(BaseWfsReader):

    """ Read density matrices in the LCAO basis from the time-dependent wave functions file.

    Yield density matrices time by time.
    """

    def __init__(self,
                 wfs_fname: str,
                 comm=world,
                 yield_re: bool = True,
                 yield_im: bool = True,
                 stridet: int = 1,
                 tmax: int = -1,
                 filter_times: list[float] | Array1D[np.float64] | None = None,
                 log: Logger | None = None,
                 striden: int = 4):

        super().__init__(wfs_fname=wfs_fname, comm=comm,
                         yield_re=yield_re, yield_im=yield_im, log=log,
                         stridet=stridet, tmax=tmax,
                         filter_times=filter_times)
        self._f_skn: NDArray[np.float64] | None = None
        self._C0_sknM: NDArray[np.float64] | None = None
        self._rho0_skMM: NDArray[np.float64] | None = None
        self._striden = striden

    @property
    def nn(self) -> int:
        return self.f_skn.shape[2]

    @property
    def striden(self) -> int:
        return self._striden

    @property
    def true_striden(self) -> int:
        if self.striden == 0:
            return self.nn
        return self.striden

    @property
    def f_skn(self) -> NDArray[np.float64]:
        """ Occupations numbers. """
        if self._f_skn is None:
            self._f_skn = proxy_occupations(self.initreader)[:]
        return self._f_skn

    @property
    def C0_sknM(self) -> NDArray[np.float64]:
        if self._C0_sknM is None:
            C0_sknM = proxy_coefficients(self.initreader)[:]
            assert np.max(np.abs(C0_sknM.imag)) < 1e-20
            self._C0_sknM = C0_sknM.real
        return self._C0_sknM

    @property
    def rho0_skMM(self) -> NDArray[np.float64]:
        if self._rho0_skMM is None:
            self.log(f'Constructing rho0_skMM on {self.comm.size} ranks',
                     who='Reader', rank=0, flush=True)
            self._rho0_skMM = calculate_rho0_parallel(self.f_skn, self.C0_sknM, comm=self.comm)
            self.log('Constructed rho0_skMM',
                     who='Reader', rank=0, flush=True)
            assert self._rho0_skMM is not None
        return self._rho0_skMM

    def inner_work_loop(self) -> Generator[slice, None, None]:
        for n in range(0, self.nn, self.true_striden):
            yield slice(n, n + self.true_striden)

    def subtract_ground_state(self,
                              dm_buffer: DensityMatrixBuffer,
                              s: int,
                              k: int,
                              M1: slice,
                              M2: slice):
        rhs = -self.rho0_skMM[s, k, M1, M2]
        dm_buffer.safe_add(True, 0, rhs)

    def nnshape(self,
                s: int,
                k: int,
                M1: slice,
                M2: slice) -> tuple[int, int]:
        M1size = M1.stop - M1.start
        M2size = M2.stop - M2.start
        MMshape = (M1size, M2size)
        return MMshape

    def iread(self,
              s: int,
              k: int,
              M1: slice,
              M2: slice) -> Generator[DensityMatrixBuffer, None, None]:
        """ Read the density matrices time by time.

        Parameters
        ----------
        s, k, M1, M2
            Read these indices.
        """
        dm_buffer = DensityMatrixBuffer(self.nnshape(s, k, M1, M2), (), np.float64)

        einsumstr = 'n,nM,nO->MO'

        self.rho0_skMM  # Construct synchronously on all ranks
        for globalt in self.work_loop(self.comm.rank):
            if globalt is None:
                continue

            if self.yield_re:
                dm_buffer.zeros(True, 0)
            if self.yield_im:
                dm_buffer.zeros(False, 0)

            reader = self.reader_t[globalt]
            for n in self.inner_work_loop():
                C_nM1 = proxy_C_nM(reader, s, k, n, M1)  # Here n is occupied states only
                C_nM2 = proxy_C_nM(reader, s, k, n, M2)
                f_n = proxy_occupations(reader, s, k)[n]
                path = np.einsum_path(einsumstr, f_n, C_nM1.real, C_nM2.real, optimize='optimal')[0]

                # Conjugate C_nM2
                if self.yield_re:
                    Rerho_x = np.einsum(einsumstr, f_n, C_nM1.real, C_nM2.real, optimize=path)
                    Rerho_x += np.einsum(einsumstr, f_n, C_nM1.imag, C_nM2.imag, optimize=path)
                    dm_buffer.safe_add(True, 0, Rerho_x)
                if self.yield_im:
                    Imrho_x = np.einsum(einsumstr, f_n, C_nM1.imag, C_nM2.real, optimize=path)
                    Imrho_x -= np.einsum(einsumstr, f_n, C_nM1.real, C_nM2.imag, optimize=path)
                    dm_buffer.safe_add(False, 0, Imrho_x)

            self.subtract_ground_state(dm_buffer, s, k, M1, M2)

            yield dm_buffer


class WfsReader(BaseWfsReader):

    """ Read wave function LCAO coefficients from the time-dependent wave functions file.

    Yield wave functions time by time.
    """

    def __init__(self,
                 wfs_fname: str,
                 comm=world,
                 yield_re: bool = True,
                 yield_im: bool = True,
                 stridet: int = 1,
                 tmax: int = -1,
                 filter_times: list[float] | Array1D[np.float64] | None = None,
                 log: Logger | None = None):
        super().__init__(wfs_fname=wfs_fname, comm=comm,
                         yield_re=yield_re, yield_im=yield_im, log=log,
                         stridet=stridet, tmax=tmax,
                         filter_times=filter_times)
        self._f_skn: NDArray[np.float64] | None = None
        self._C0_sknM: NDArray[np.float64] | None = None

    @property
    def nn(self) -> int:
        return self.C0_sknM.shape[2]

    @property
    def nM(self) -> int:
        return self.C0_sknM.shape[3]

    @property
    def f_skn(self) -> NDArray[np.float64]:
        """ Occupations """
        if self._f_skn is None:
            self._f_skn = proxy_occupations(self.initreader)[:]
        return self._f_skn

    @property
    def C0_sknM(self) -> NDArray[np.float64]:
        if self._C0_sknM is None:
            C0_sknM = proxy_coefficients(self.initreader)[:]
            assert np.max(np.abs(C0_sknM.imag)) < 1e-20
            self._C0_sknM = C0_sknM.real
        return self._C0_sknM

    def nnshape(self,
                s: int,
                k: int,
                n: slice,
                M: slice) -> tuple[int, int]:
        nsize = n.stop - n.start
        Msize = M.stop - M.start
        nMshape = (nsize, Msize)
        return nMshape

    def iread(self,
              s: int,
              k: int,
              n: slice,
              M: slice) -> Generator[DensityMatrixBuffer, None, None]:
        """ Read the density matrices time by time.

        Parameters
        ----------
        s, k, n, M
            Read these indices.
        """
        dm_buffer = DensityMatrixBuffer(self.nnshape(s, k, n, M), (), np.float64)

        for globalt in self.work_loop(self.comm.rank):
            if globalt is None:
                continue

            if self.yield_re:
                dm_buffer.zeros(True, 0)
            if self.yield_im:
                dm_buffer.zeros(False, 0)

            reader = self.reader_t[globalt]
            C_nM = proxy_C_nM(reader, s, k, n, M)
            if self.yield_re:
                dm_buffer.safe_fill(True, 0, C_nM.real)
            if self.yield_im:
                dm_buffer.safe_fill(False, 0, C_nM.imag)

            yield dm_buffer


def prepare_wave_function_readers(mainreader,
                                  comm,
                                  log: Callable = print,
                                  stridet: int = 1,
                                  tmax: int = -1,
                                  parallel: bool = True,
                                  ) -> tuple[Array1D[np.float64], WaveFunctionReader, list[WaveFunctionReader]]:
    readerlen = len(mainreader)
    log(f'Opening time-dependent wave functions file {mainreader.filename}', who='Reader', rank=0, flush=True)

    # Encode the action as int according to the following list
    int2action = [None, 'init', 'kick', 'propagate']
    action2int = {a: i for i, a in enumerate(int2action)}

    if parallel:
        # Read in parallel
        readrange = range(0, readerlen, comm.size)
    else:
        # Read everything on root
        readrange = range(readerlen) if comm.rank == 0 else range(0)

    if comm.rank == 0 or parallel:
        # Buffers to read to on all ranks when parallel
        time_1 = np.array([0], dtype=float)
        action_1 = np.array([0], dtype=int)

    def round_up(val):
        v = (val + comm.size - 1)
        v = v // comm.size
        return v * comm.size

    if comm.rank == 0:
        # Buffers for gathering
        time_t = np.zeros(round_up(readerlen), dtype=time_1.dtype)
        action_t = np.zeros(round_up(readerlen), dtype=action_1.dtype)

    # Read all times in parallel, gather on root
    for roott in readrange:
        t = roott + comm.rank
        if t < readerlen:
            log(f'Opened item #{t}', who='Reader', comm=comm, if_elapsed=5, flush=True)

            reader = mainreader[t]
            action = getattr(reader, 'action', None)
            action_1[:] = action2int[action]
            try:
                time_1[:] = getattr(reader, 'time')
            except AttributeError:
                # Depending on the action these might not be set
                time_1[:] = np.nan

        if parallel:
            comm.gather(time_1, 0, time_t[roott:roott + comm.size] if comm.rank == 0 else None)
            comm.gather(action_1, 0, action_t[roott:roott + comm.size] if comm.rank == 0 else None)
        else:
            time_t[t] = time_1[0]
            action_t[t] = action_1[0]

    if comm.rank == 0:
        # The root rank must count the number of time entries
        nreadt = 0
        lasttime = -np.inf

        readtimes = []
        timereadert = []
        initreadert = None
        for t in range(readerlen):
            action = int2action[action_t[t]]
            curtime = time_t[t]

            if curtime <= lasttime:
                log(f'Negative time difference at #{t} '
                    f'({curtime:.1f} <= {lasttime:.1f}). Skipping',
                    who='Reader', comm=comm, flush=True)
                continue

            if action is None:
                # This is just some dummy entry
                continue

            # Find the first 'init' entry
            if action == 'init':
                assert curtime >= lasttime, f'Times not matching {t}:{curtime} !>= {lasttime}'
                if initreadert is None:
                    initreadert = t
                continue

            assert action in ['kick', 'propagate']
            assert curtime != np.nan
            if nreadt % stridet == 0:
                # Save multiples of stridet
                # Note that 0 will always be saved. That contains the kick
                readtimes.append(curtime)
                timereadert.append(t)
            nreadt += 1
            if tmax > 0 and nreadt >= tmax:
                break
            lasttime = curtime

        assert initreadert is not None
        broadcast_obj = (initreadert, timereadert, np.array(readtimes))
    else:
        broadcast_obj = None

    # Distribute
    (initreadert, timereadert, time_t) = broadcast(broadcast_obj, comm=comm, root=0)
    initreader = mainreader[initreadert]
    timereader_t = [mainreader[t] for t in timereadert]

    log(f'Opened time-dependent wave functions file with {len(timereader_t)} times',
        who='Reader', rank=0, flush=True)

    return time_t, initreader, timereader_t


def read_C0S_parallel(ksdreader: Reader,
                      comm=None) -> NDArray[np.float64]:
    if comm is None:
        comm = world

    # Compute C0S in parallel
    C0_sknM = ksdreader.proxy('C0_unM', 0)
    S_skMM = ksdreader.proxy('S_uMM', 0)
    nM = C0_sknM.shape[3]

    # Rank local contributions
    strideM = (nM + comm.size - 1) // comm.size
    sliceM = slice(strideM * comm.rank, strideM * (comm.rank + 1))
    C0_sknM = C0_sknM[:][..., sliceM]
    S_skMM = S_skMM[:][..., sliceM, :]

    # Compute C0S
    C0S_sknM = np.einsum('sknO,skOM->sknM', C0_sknM, S_skMM, optimize=True)

    # Sum and distribute
    comm.sum(C0S_sknM)

    return C0S_sknM


def calculate_rho0_parallel(f_skn: NDArray[np.float64],
                            C0_sknM: NDArray[np.float64],
                            comm=None) -> NDArray[np.float64]:
    if comm is None:
        comm = world

    nn = f_skn.shape[2]
    striden = (nn + comm.size - 1) // comm.size
    slicen = slice(striden * comm.rank, striden * (comm.rank + 1))
    f_skn = f_skn[:][..., slicen]
    C0_sknM = C0_sknM[:][..., slicen, :]

    # Compute density matrix
    rho0_skMM = np.einsum('skn,sknM,sknO->skMO',
                          f_skn, C0_sknM, C0_sknM,
                          optimize=True)

    # Sum and distribute
    comm.sum(rho0_skMM)

    return rho0_skMM


def proxy_coefficients(reader, *indices):
    """ Proxy the wave function coefficients, with the correct units."""
    coefficients = reader.wave_functions.proxy('coefficients', *indices)
    coefficients.scale = Bohr ** 1.5

    return coefficients


def proxy_occupations(reader, *indices):
    """ Proxy the wave function occupations with the correct scale."""
    occupations = reader.wave_functions.proxy('occupations', *indices)
    occupations.scale = 2 / reader.wave_functions.occupations.shape[0]

    return occupations


def proxy_C_nM(reader,
               *indices):
    """ Proxy the wave function coefficients, with the correct units.

    Make sure that the proxied array has at least two dimensions"""
    x = indices[:-2]
    n = indices[-2]
    M = indices[-1]

    C_nM = proxy_coefficients(reader, *x)
    if isinstance(n, slice):
        C_x = C_nM[n][:, M]
    else:
        C_M = C_nM.proxy(n)
        C_x = np.atleast_2d(C_M[M])
    assert len(C_x.shape) == 2
    return C_x



from __future__ import annotations

from typing import Any
from pathlib import Path
import numpy as np
from numpy.typing import NDArray

from ase.parallel import parprint
from gpaw.mpi import world
from gpaw.tddft.units import as_to_au, au_to_as, eV_to_au, au_to_eV, eA_to_au, au_to_eA
from gpaw.tddft.spectrum import read_dipole_moment_file

from .perturbation import create_perturbation, Perturbation, PerturbationLike, NoPerturbation


class SpectrumCalculator:

    r""" Spectrum calculator.

    Calculates the dynamic polarizability and dipole strength function (the spectrum).

    The polarizability :math:`\alpha(\omega)` is related to the perturbation
    :math:`\mathcal{E}(\omega)` and the Fouier transform of the dipole moment
    :math:`\boldsymbol\mu(t)` as

    .. math::

        \boldsymbol\alpha(\omega) \mathcal{E}(\omega) = \mathcal{F}[\boldsymbol\mu(t)](\omega).

    The dipole strength function is

    .. math::

        \boldsymbol{S}(\omega) = \frac{2}{\pi} \omega\:\mathrm{Im}[\boldsymbol\alpha(\omega)].

    Both quantities can be broadened by supplying a non-zero value :math:`\sigma`. Then the
    convolution

    .. math::

        \frac{1}{\sqrt{2\pi/\sigma^2}}
        \int_{-\infty}^{\infty} \boldsymbol\alpha(\omega')
        \exp\left(- \frac{(\omega - \omega')^2}{2\sigma^2}\right) \mathrm{d}\omega'

    is computed. When the perturbation is a delta-kick, this can efficiently be computed
    by multiplying the dipole moment by a Gaussian envelope before Fourier transformation

    .. math::

        \boldsymbol\mu(t) \exp(-\sigma^2 t^2 / 2).

    For other perturbations, a :term:`FFT` and :term:`IFFT` is first
    performed to obtain the delta-kick response.


    Parameters
    ----------
    times
        Array (length T) of times in units of as.
    dipole_moments
        Array (shape (T, 3)) of dipole moments in units of eÅ.
    perturbation
        The perturbation that was applied in the :term:`TDDFT` calculation.
    """

    def __init__(self,
                 times: NDArray[np.float64],
                 dipole_moments: NDArray[np.float64],
                 perturbation: PerturbationLike):
        time_t = np.array(times) * as_to_au
        dm_tv = np.array(dipole_moments) * eA_to_au
        dm_tv -= dm_tv[0]  # Remove static dipole

        # Remove duplicates due to stopped and restarted calculations, and delta kick
        flt_t = np.ones_like(time_t, dtype=bool)
        maxtime = time_t[0]
        for t in range(1, len(time_t)):
            if time_t[t] > maxtime:
                maxtime = time_t[t]
            else:
                flt_t[t] = False
        time_t = time_t[flt_t]
        dm_tv = dm_tv[flt_t]
        ndup = len(flt_t) - flt_t.sum()
        if ndup > 0:
            parprint(f'Removed {ndup} duplicates')

        # check time step
        dt_t = time_t[1:] - time_t[:-1]
        dt = dt_t[0]
        if not np.allclose(dt_t, dt, rtol=1e-6, atol=0):
            raise ValueError('Time grid must be equally spaced.')

        self._time_t = time_t
        self._dm_tv = dm_tv
        self._perturbation = create_perturbation(perturbation)

        if isinstance(self.perturbation, NoPerturbation):
            raise ValueError('A perturbation must be given')

    @property
    def times(self) -> NDArray[np.float64]:
        """ Array of times corresponding to dipole moments, in units of as. """
        return self._time_t * au_to_as

    @property
    def dipole_moments(self) -> NDArray[np.float64]:
        """ Array of dipole moments, in units of eÅ. """
        return self._dm_tv * au_to_eA

    @property
    def perturbation(self) -> Perturbation:
        """ The perturbation that was applied in the :term:`TDDFT` calculation. """
        return self._perturbation

    def _get_dynamic_polarizability(self,
                                    frequencies: list[float] | NDArray[np.float64],
                                    frequency_broadening: float = 0):
        """ Calculate the dynamic polarizability in atomic units.

        Parameters
        ----------
        frequencies
            Array of frequencies for which to calculate the polarizability; in atomic units.
        frequency_broadening
            Gaussian broadening width in atomic units. Default (0) is no broadening.

        Returns
        -------
        Array of dynamic polarizabilities in atomic units. \
        The array has shape (N, 3), where N is the length of :attr:`frequencies`.
        """
        dt = self._time_t[1] - self._time_t[0]
        nt = len(self._time_t)

        # Get a response equivalent to a unity-strength delta-kick
        dm_tv = self.perturbation.normalize_time_response(self._dm_tv, self._time_t, axis=0)

        if frequency_broadening == 0:
            # No broadening
            weight_t = np.ones_like(self._time_t)
        else:
            # Gaussian broadening
            weight_t = np.exp(-0.5 * frequency_broadening ** 2 * self._time_t**2)

        # integration weights from Simpson's integration rule
        weight_t *= dt / 3 * np.array([1] + [4, 2] * int((nt - 2) / 2)
                                      + [4] * (nt % 2) + [1])

        # transform
        exp_tw = np.exp(np.outer(1.0j * self._time_t, frequencies))
        dm_wv = np.einsum('t...,tw,t->w...', dm_tv, exp_tw, weight_t, optimize=True)

        return dm_wv

    def get_dynamic_polarizability(self,
                                   frequencies: list[float] | NDArray[np.float64],
                                   frequency_broadening: float = 0):
        """ Calculate the dynamic polarizability.

        Parameters
        ----------
        frequencies
            Array of frequencies for which to calculate the polarizability; in units of eV.
        frequency_broadening
            Gaussian broadening width in atomic units. Default (0) is no broadening.

        Returns
        -------
        Array of dynamic polarizabilities in (eÅ)**2/eV. \
        The array has shape (N, 3), where N is the length of :attr:`frequencies`.
        """

        dm_wv = self._get_dynamic_polarizability(np.array(frequencies) * eV_to_au,
                                                 frequency_broadening * eV_to_au)
        dm_wv = dm_wv * au_to_eA**2 / au_to_eV
        return dm_wv

    def _get_dipole_strength_function(self,
                                      frequencies: list[float] | NDArray[np.float64],
                                      frequency_broadening: float = 0):
        """ Calculate the dipole strength function (spectrum) in atomic units.

        Parameters
        ----------
        frequencies
            Array of frequencies for which to calculate the spectrum; in atomic units.
        frequency_broadening
            Gaussian broadening width in atomic units. Default (0) is no broadening.

        Returns
        -------
        Array of dipole strength function in atomic units. \
        The array has shape (N, 3), where N is length of frequencies.
        """
        freq_w = np.array(frequencies)
        pol_wv = self._get_dynamic_polarizability(frequencies, frequency_broadening)
        osc_wv = 2 / np.pi * freq_w[:, np.newaxis] * pol_wv.imag
        return osc_wv  # type: ignore

    def get_dipole_strength_function(self,
                                     frequencies: list[float] | NDArray[np.float64],
                                     frequency_broadening: float = 0):
        """ Calculate the dipole strength function (spectrum) in units of 1/eV.

        Parameters
        ----------
        frequencies
            Array of frequencies for which to calculate the spectrum; in units of eV.
        frequency_broadening
            Gaussian broadening width in units of eV. Default (0) is no broadening.

        Returns
        -------
        Array of dipole strength function in units of 1/eV. \
        The array has shape (N, 3), where N is length of frequencies.
        """
        osc_wv = self._get_dipole_strength_function(np.array(frequencies) * eV_to_au,
                                                    frequency_broadening * eV_to_au)
        osc_wv = osc_wv / au_to_eV
        return osc_wv

    @classmethod
    def from_file(cls,
                  dipolefile: str | Path,
                  perturbation: PerturbationLike) -> SpectrumCalculator:
        _, time_t, _, dm_tv = read_dipole_moment_file(str(dipolefile))
        return cls(time_t * au_to_as, dm_tv * au_to_eA, perturbation)

    def calculate_spectrum_and_write(self,
                                     out_fname: Path | str,
                                     frequencies: list[float] | NDArray[np.float64],
                                     frequency_broadening: float = 0,
                                     write_extra: dict[str, Any] = dict()):
        """ Calculate the dipole strength function (spectrum) and write to file.

        The spectrum is saved in a numpy archive if the file extension is `.npz`
        or in a comma separated file if the file extension is `.dat`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        frequencies
            Array of frequencies for which to calculate the spectrum; in units of eV.
        frequency_broadening
            Gaussian broadening width in units of eV. Default (0) is no broadening.
        write_extra
            Dictionary of additional data written to numpy archive (ignored for `.dat`) files.
        """
        from .writers.spectrum import write_spectrum

        out_fname = str(out_fname)
        if world.rank > 0:
            world.barrier()
            return

        osc_wv = self.get_dipole_strength_function(frequencies, frequency_broadening)
        if out_fname[-4:] == '.npz':
            d = dict(freq_w=frequencies,
                     osc_wv=osc_wv,
                     frequency_broadening=frequency_broadening)
            d.update({f'perturbation_{key}': value
                      for key, value in self.perturbation.todict().items()})
            d.update(write_extra)
            np.savez_compressed(str(out_fname), **d)
        elif out_fname[-4:] == '.dat':
            write_spectrum(out_fname,
                           frequencies=frequencies,
                           spectrum=osc_wv,
                           frequency_broadening=frequency_broadening,
                           total_time=self.times[-1] - self.times[0],
                           timestep=self.times[1] - self.times[0],
                           perturbation=self.perturbation)
        else:
            raise ValueError(f'output-file must have ending .npz or .dat, is {out_fname}')

        print(f'Written {out_fname}', flush=True)
        world.barrier()



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from gpaw.mpi import world

from ..voronoi import AtomProjectionsType


def write_density_of_states(out_fname: str,
                            energies: list[float] | NDArray[np.float64],
                            dos: list[float] | NDArray[np.float64],
                            sigma: float,
                            zerofermi: bool = False):
    """ Write the broadened :term:`DOS` to a text file.

    Parameters
    ----------
    out_fname
        File name of the resulting data file.
    energies
        Array of energies in units of eV.
    dos
        Array of DOS corresponding to the :attr:`energies`.
    sigma
        Gaussian broadening width in units of eV.
    zerofermi
        True if energies are to be relative to Fermi level, False if relative to vacuum.
    """
    if zerofermi:
        zerostr = 'relative to Fermi level'
    else:
        zerostr = 'relative to vacuum level'

    header = (f'DOS {zerostr}\n'
              f'Gaussian folding, Width {sigma:.4f} eV\n'
              'Energy (eV)        DOS (1/eV)')
    np.savetxt(out_fname, np.array([energies, dos]).T, fmt=['%12.6f', '%18.8e'], header=header)


def write_partial_density_of_states(out_fname: str,
                                    energies: list[float] | NDArray[np.float64],
                                    pdos: list[float] | NDArray[np.float64],
                                    atom_projections: AtomProjectionsType,
                                    sigma: float,
                                    zerofermi: bool = False):
    """ Write the broadened :term:`PDOS` to a text file.

    Parameters
    ----------
    out_fname
        File name of the resulting data file.
    energies
        Array of energies in units of eV.
    pdos
        Array of PDOS corresponding to the :attr:`energies`.
    atom_projections
        Atom projections.
    sigma
        Gaussian broadening width in units of eV.
    zerofermi
        True if energies are to be relative to Fermi level, False if relative to vacuum.
    """
    if world.rank != 0:
        return

    Ni = len(atom_projections)
    if zerofermi:
        zerostr = 'relative to Fermi level'
    else:
        zerostr = 'relative to vacuum level'

    savedata = np.zeros((len(energies), Ni + 1))
    savedata[:, 0] = energies
    savedata[:, 1:] = pdos

    projectionsstr = '\n'.join([f'  {i:4.0f}: {str(proj)}'
                                for i, proj in enumerate(atom_projections)])
    projcolumns = '   '.join([f'PDOS {i:4.0f} (1/eV)' for i in range(Ni)])

    header = (f'PDOS {zerostr}\n'
              'Atomic projections:\n'
              f'{projectionsstr}\n'
              f'Gaussian folding, Width {sigma:.4f} eV\n'
              f'Energy (eV)   {projcolumns}')
    fmt = ['%13.6f'] + Ni*['%18.8e']
    np.savetxt(out_fname, savedata, fmt, header=header)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy as np
from numpy._typing import _DTypeLike as DTypeLike


@dataclass
class MemoryEntry:

    shape: tuple[int, ...]
    dtype: np.dtype
    on_num_ranks: int = 1
    total_size: int | None = None

    def get_total_size(self) -> int:
        """ Get the total number of elements on all ranks. """
        if self.total_size is not None:
            return self.total_size
        return int(np.prod(self.shape)) * self.on_num_ranks


@dataclass
class MemoryEstimate:

    comment: str = ''
    children: dict[str, MemoryEstimate] = field(default_factory=dict)
    arrays: dict[str, MemoryEntry] = field(default_factory=dict)

    def __str__(self) -> str:
        if len(self.children) == 0 and len(self.arrays) == 0:
            return 'Unknown'
        to_MiB = 1024 ** -2
        totalstr = f'{self.grand_total * to_MiB:.1f} MiB'

        lines = []
        if self.comment != '':
            lines += ['Note: ' + line for line in self.comment.split('\n')]
            lines.append('')

        for key, entry in self.arrays.items():
            size_per_rank_MiB = int(np.prod(entry.shape)) * entry.dtype.itemsize * to_MiB
            size_total_MiB = entry.get_total_size() * entry.dtype.itemsize * to_MiB

            lines.append(f'{key}: {entry.shape} {entry.dtype}')
            lines.append(f'. {size_per_rank_MiB:.1f} MiB '
                         f'per rank on {entry.on_num_ranks} ranks')
            lines.append(f'. {size_total_MiB:.1f} MiB in total on all ranks')
            lines.append('')

        for name, child in self.children.items():
            lines.append(f'{name}:')
            lines += ['    ' + line for line in str(child).split('\n')]
            lines.append('')
        lines.append(f'{"":.^24}')
        lines.append(f'{" Total on all ranks ":.^24}')
        lines.append(f'{totalstr:.^24}')

        return '\n'.join(lines)

    @property
    def grand_total(self) -> int:
        """ Grand total of bytes. """

        total = 0
        for entry in self.arrays.values():
            total += entry.get_total_size() * entry.dtype.itemsize

        for child in self.children.values():
            total += child.grand_total

        return total

    def add_key(self,
                key: str,
                shape: tuple[int, ...] | int = (),
                dtype: DTypeLike = float,
                *,
                total_size: int | None = None,
                on_num_ranks: int = 1):
        assert isinstance(key, str)
        if isinstance(shape, int):
            shape = (shape, )
        assert all([isinstance(d, (int, np.integer)) for d in shape])
        shape = tuple(int(d) for d in shape)
        assert isinstance(on_num_ranks, int)
        dtype = np.dtype(dtype)
        self.arrays[key] = MemoryEntry(shape, dtype,
                                       total_size=total_size,
                                       on_num_ranks=on_num_ranks)

    def add_child(self,
                  name: str,
                  child: MemoryEstimate):
        assert isinstance(child, MemoryEstimate)
        self.children[name] = child


class HasMemoryEstimate(ABC):

    """ Classes inheriting from this class are able to
    provide a memory estimate """

    @abstractmethod
    def get_memory_estimate(self) -> MemoryEstimate:
        raise NotImplementedError



from __future__ import annotations

import re
from contextlib import nullcontext
from operator import itemgetter
from pathlib import Path
from typing import Any, Callable, Generic, Iterable, NamedTuple, TypeVar
import numpy as np
from numpy.typing import NDArray
from numpy._typing import _DTypeLike as DTypeLike  # parametrizable wrt generic

from ase.io.ulm import open
from ase.parallel import parprint
from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition
from gpaw.lcaotddft.laser import GaussianPulse
from gpaw.mpi import SerialCommunicator, world
from gpaw.tddft.units import fs_to_au, au_to_eV

from .logging import Logger, NoLogger
from .result import Result, ResultKeys
from ..perturbation import Perturbation
from ..typing import Array1D, Communicator

__all__ = [
    'Logger',
    'Result',
    'ResultKeys',
]


DTypeT = TypeVar('DTypeT', bound=np.generic, covariant=True)


class ParallelMatrix(Generic[DTypeT]):

    """ Distributed array, with data on the root rank.

    Parameters
    ----------
    shape
        Shape of array.
    dtype
        Dtype of array.
    comm
        MPI communicator.
    array
        Array on root rank of the communicator. Must be ``None`` on other ranks.
    """

    def __init__(self,
                 shape: tuple[int, ...],
                 dtype: DTypeLike[DTypeT],
                 comm: Communicator | None = None,
                 array: NDArray[DTypeT] | None = None):
        if comm is None:
            comm = world
        self.comm = comm
        self.shape = shape
        self.dtype = np.dtype(dtype)

        self._array: NDArray[DTypeT] | None
        if self.root:
            assert array is not None
            assert array.shape == shape
            assert array.dtype == np.dtype(dtype)
            self._array = array
        else:
            assert array is None
            self._array = None

    @property
    def array(self) -> NDArray[DTypeT]:
        """ Array with data. May only be called on the root rank. """
        if not self.root:
            raise RuntimeError('May only be called on root')
        assert self._array is not None
        return self._array

    @property
    def root(self) -> bool:
        """ Whether this rank is the root rank. """
        return self.comm.rank == 0

    @property
    def T(self) -> ParallelMatrix:
        shape = self.shape[:-2] + self.shape[-2:][::-1]
        return ParallelMatrix(shape=shape, dtype=self.dtype, comm=self.comm,
                              array=self.array.T if self.root else None)

    def broadcast(self) -> NDArray[DTypeT]:
        """ Broadcasted data. """
        if self.root:
            A = np.ascontiguousarray(self.array)
        else:
            A = np.zeros(self.shape, self.dtype)

        self.comm.broadcast(A, 0)

        return A

    def __matmul__(self, other) -> ParallelMatrix[DTypeT]:
        """ Perform matrix multiplication in parallel. """
        if not isinstance(other, ParallelMatrix):
            raise NotImplementedError

        assert self.dtype == other.dtype

        A = self.broadcast()
        B = other.broadcast()

        # Allocate array for result
        ni, nj = A.shape[-2:]
        nk = B.shape[-1]
        C_shape = np.broadcast_shapes(A.shape[:-2], B.shape[:-2]) + (ni, nk)
        C = np.zeros(C_shape, self.dtype)

        # Determine slice for ranks
        stridek = (nk + self.comm.size - 1) // self.comm.size
        slicek = slice(stridek * self.comm.rank, stridek * (self.comm.rank + 1))

        # Perform the matrix multiplication
        C[..., :, slicek] = A @ B[..., :, slicek]

        # Sum to root rank
        self.comm.sum(C, 0)

        result = ParallelMatrix(C_shape, self.dtype, comm=self.comm,
                                array=C if self.root else None)
        return result


def gauss_ij_with_filter(energy_i: np.typing.ArrayLike,
                         energy_j: np.typing.ArrayLike,
                         sigma: float,
                         fltthresh: float | None = None,
                         ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    r""" Computes the matrix

    .. math::

        M_{ij}
        = \frac{1}{\sqrt{2 \pi \sigma^2}}
        \exp\left(-\frac{
            \left(\varepsilon_i - \varepsilon_j\right)^2
        }{
            2 \sigma^2
        }\right)

    Useful for Gaussian broadening. Optionally only computes the exponent
    above a certain threshold, and returns the filter.

    Parameters
    ----------
    energy_i
        Energies :math:`\varepsilon_i`.
    energy_j
        Energies :math:`\varepsilon_j`.
    sigma
        Gaussian broadening width :math:`\sigma`.
    fltthresh
        Filtering threshold.

    Returns
    -------
        Matrix :math:`M_{ij}`, filter.
    """
    energy_i = np.asarray(energy_i)
    energy_j = np.asarray(energy_j)

    norm = 1.0 / (sigma * np.sqrt(2 * np.pi))

    denergy_ij = energy_i[:, np.newaxis] - energy_j[np.newaxis, :]
    exponent_ij = -0.5 * (denergy_ij / sigma) ** 2

    if fltthresh is not None:
        flt_i = np.any(exponent_ij > fltthresh, axis=1)
        M_ij = np.zeros_like(exponent_ij)
        M_ij[flt_i] = norm * np.exp(exponent_ij[flt_i])
    else:
        flt_i = np.ones(energy_i.shape, dtype=bool)
        M_ij = norm * np.exp(exponent_ij)

    return M_ij, flt_i  # type: ignore


def gauss_ij(energy_i: np.typing.ArrayLike,
             energy_j: np.typing.ArrayLike,
             sigma: float) -> NDArray[np.float64]:
    r""" Computes the matrix

    .. math::

        M_{ij}
        = \frac{1}{\sqrt{2 \pi \sigma^2}}
        \exp\left(-\frac{
            \left(\varepsilon_i - \varepsilon_j\right)^2
        }{
            2 \sigma^2
        }\right),

    which is useful for Gaussian broadening.

    Parameters
    ----------
    energy_i
        Energies :math:`\varepsilon_i`.
    energy_j
        Energies :math:`\varepsilon_j`.
    sigma
        Gaussian broadening width :math:`\sigma`.

    Returns
    -------
        Matrix :math:`M_{ij}`.
    """
    M_ij, _ = gauss_ij_with_filter(energy_i, energy_j, sigma)
    return M_ij


def broaden_n2e(M_n: np.typing.ArrayLike,
                energy_n: np.typing.ArrayLike,
                energy_e: np.typing.ArrayLike,
                sigma: float) -> NDArray[np.float64]:
    r""" Broaden matrix onto energy grids

    .. math::

        M(\varepsilon_e)
        = \sum_n M_n \frac{1}{\sqrt{2 \pi \sigma^2}}
        \exp\left(-\frac{
            \left(\varepsilon_n - \varepsilon_e\right)^2
        }{
            2 \sigma^2
        }\right),

    Returns
    -------
        :math:`M(\varepsilon_0)`
    """
    M_n = np.asarray(M_n)
    gauss_ne, flt_n = gauss_ij_with_filter(energy_n, energy_e, sigma)

    M_e = np.einsum('n,ne->e', M_n[flt_n], gauss_ne[flt_n], optimize=True)

    return M_e


def broaden_xn2e(M_xn: np.typing.ArrayLike,
                 energy_n: np.typing.ArrayLike,
                 energy_e: np.typing.ArrayLike,
                 sigma: float) -> NDArray[np.float64]:
    r""" Broaden matrix onto energy grids

    .. math::

        M(\varepsilon_e)
        = \sum_n M_n \frac{1}{\sqrt{2 \pi \sigma^2}}
        \exp\left(-\frac{
            \left(\varepsilon_n - \varepsilon_e\right)^2
        }{
            2 \sigma^2
        }\right).

    Returns
    -------
        :math:`M(\varepsilon_0)`.
    """
    M_xn = np.asarray(M_xn)
    gauss_ne, flt_n = gauss_ij_with_filter(energy_n, energy_e, sigma)

    M_xe = np.einsum('xn,ne->xe',
                     M_xn.reshape((-1, len(flt_n)))[:, flt_n],
                     gauss_ne[flt_n],
                     optimize=True).reshape(M_xn.shape[:-1] + (-1, ))

    return M_xe


def broaden_ia2ou(M_ia: np.typing.ArrayLike,
                  energy_i: np.typing.ArrayLike,
                  energy_a: np.typing.ArrayLike,
                  energy_o: np.typing.ArrayLike,
                  energy_u: np.typing.ArrayLike,
                  sigma: float) -> NDArray[np.float64]:
    r""" Broaden matrix onto energy grids.

    .. math::

        M(\varepsilon_o, \varepsilon_u)
        = \sum_{ia} M_{ia} \frac{1}{\sqrt{2 \pi \sigma^2}}
        \exp\left(-\frac{
            (\varepsilon_i - \varepsilon_o)^2
        }{
            2 \sigma^2
        }\right)
        \exp\left(-\frac{
            (\varepsilon_a - \varepsilon_u)^2
        }{
            2 \sigma^2
        }\right)

    Returns
    -------
        :math:`M(\varepsilon_o, \varepsilon_u)`.
    """
    M_ia = np.asarray(M_ia)
    ia_shape = M_ia.shape[:2]
    x_shape = M_ia.shape[2:]
    M_iax = M_ia.reshape(ia_shape + (-1, ))
    gauss_io, flt_i = gauss_ij_with_filter(energy_i, energy_o, sigma)
    gauss_au, flt_a = gauss_ij_with_filter(energy_a, energy_u, sigma)

    M_oux = np.einsum('iax,io,au->oux', M_iax[flt_i, :][:, flt_a],
                      gauss_io[flt_i], gauss_au[flt_a],
                      optimize=True, order='C')

    return M_oux.reshape(M_oux.shape[:2] + x_shape)


def get_array_filter(values: Array1D[np.float64] | list[float],
                     filter_values: Array1D[np.float64] | list[float] | None,
                     ) -> slice | Array1D[np.bool_]:
    """ Get array filter that can be used to filter out data.

    Parameters
    ----------
    values
        Array of values, e.g. linspace of times or frequencies.
    filter_values
        List of values that one wishes to extract. The closes values from values
        will be selected as filter.

    Returns
    -------
    Object that can be used to index values and arrays with the same shape as values.
    """
    _values = np.array(values)
    flt_x: slice | NDArray[np.bool_]
    if len(values) == 0:
        # Empty list of arrays
        return slice(None)

    if filter_values is None or len(filter_values) == 0:
        # No filter
        return slice(None)

    flt_x = np.zeros(len(values), dtype=bool)
    for filter_value in filter_values:
        # Search for closest value
        idx = np.argmin(np.abs(_values - filter_value))
        flt_x[idx] = True

    return flt_x


def filter_array(values: Array1D[np.float64] | list[float],
                 filter_values: Array1D[np.float64] | list[float] | None,
                 ) -> Array1D[np.float64]:
    """ Filter array, picking values closest to :attr:`filter_values`.

    Parameters
    ----------
    values
        Array of values, e.g. linspace of times or frequencies.
    filter_values
        List of values that one wishes to extract. The closes values from values
        will be selected as filter.

    Returns
    -------
    Filtered array.
    """
    array = np.array(values)
    return array[get_array_filter(array, filter_values)]  # type: ignore


def two_communicator_sizes(*comm_sizes) -> tuple[int, int]:
    assert len(comm_sizes) == 2
    comm_size_c: list[int] = [world.size if size == 'world' else size for size in comm_sizes]
    if comm_size_c[0] == -1:
        comm_size_c[0] = world.size // comm_size_c[1]
    elif comm_size_c[1] == -1:
        comm_size_c[1] = world.size // comm_size_c[0]

    assert np.prod(comm_size_c) == world.size, \
        f'Communicator sizes must factorize world size {world.size} '\
        'but they are ' + ' and '.join([str(s) for s in comm_size_c]) + '.'
    return comm_size_c[0], comm_size_c[1]


def two_communicators(*comm_sizes) -> tuple[Communicator, Communicator]:
    """ Create two MPI communicators.

    Must satisfy ``comm_sizes[0] * comm_sizes[1] = world.size``.

    The second communicator has the ranks in sequence.

    Example
    -------

    >>> world.size == 8
    >>> two_communicators(2, 4)

    This gives::

        [0, 4]
        [1, 5]
        [2, 6]
        [3, 7]

    and::

        [0, 1, 2, 3]
        [4, 5, 6, 7]
    """
    comm_size_c = two_communicator_sizes(*comm_sizes)

    # Create communicators
    if comm_size_c[0] == 1:
        return (SerialCommunicator(), world)  # type: ignore
    elif comm_size_c[0] == world.size:
        return (world, SerialCommunicator())  # type: ignore
    else:
        assert world.size % comm_size_c[0] == 0, world.size
        # Comm 2, ranks in sequence. Comm 1, ranks skip by size of comm 2
        first_rank_in_comm_c = [world.rank % comm_size_c[1],
                                world.rank - world.rank % comm_size_c[1]]
        step_c = [comm_size_c[1], 1]
        comm_ranks_cr = [list(range(start, start + size*step, step))
                         for start, size, step in zip(first_rank_in_comm_c, comm_size_c, step_c)]
        comm_c = [world.new_communicator(comm_ranks_r) for comm_ranks_r in comm_ranks_cr]
        return comm_c[0], comm_c[1]


def detect_repeatrange(n: int,
                       stride: int,
                       verbose: bool = True) -> slice | None:
    """ If an array of length :attr:`n` is not divisible by the stride :attr:`stride`
    then some work will have to be repeated
    """
    final_start = (n // stride) * stride
    repeatrange = slice(final_start, n)
    if repeatrange.start == repeatrange.stop:
        return None
    else:
        print(f'Detected repeatrange {repeatrange}', flush=True)
        return repeatrange

    return None


def safe_fill(a: NDArray[DTypeT],
              b: NDArray[DTypeT]):
    """ Perform the operation ``a[:] = b``, checking if the dimensions match.

    If the dimensions of :attr:`b` are larger than the dimensions of :attr:`a`, raise an error.

    If the dimensions of :attr:`b` are smaller than the dimensions of :attr:`a`, write to
    the first elements of :attr:`a`.
    """
    assert len(a.shape) == len(b.shape), f'{a.shape} != {b.shape}'
    assert all([dima >= dimb for dima, dimb in zip(a.shape, b.shape)]), f'{a.shape} < {b.shape}'
    s = tuple([slice(dim) for dim in b.shape])
    a[s] = b


def safe_fill_larger(a: NDArray[DTypeT],
                     b: NDArray[DTypeT]):
    """ Perform the operation ``a[:] = b``, checking if the dimensions match.

    If the dimensions of :attr:`b` are smaller than the dimensions of :attr:`a`, raise an error.

    If the dimensions of :attr:`b` are larger than the dimensions of :attr:`a`, write the first
    elements of :attr:`b` to :attr:`a`.
    """
    assert len(a.shape) == len(b.shape), f'{a.shape} != {b.shape}'
    assert all([dimb >= dima for dima, dimb in zip(a.shape, b.shape)]), f'{a.shape} > {b.shape}'
    s = tuple([slice(dim) for dim in a.shape])
    a[:] = b[s]


IND = TypeVar('IND', slice, tuple[slice, ...])


def concatenate_indices(indices_list: Iterable[IND],
                        ) -> tuple[IND, list[IND]]:
    """ Concatenate indices.

    Given an array A and a list of incides indices_list such that A can be indexed

    >>> for indices in indices_list:
    >>>     A[indices]

    this function shall concatenate the indices into indices_concat so that the array
    can be indexed in one go. This function will also give a new list of indices
    new_indices_list that can be used to index the ``A[indices_concat]``. The following
    snippet shall be equivalent to the previous snipped.

    >>> B = A[indices_concat]
    >>> for indices in new_indices_list:
    >>>     B[indices]

    Note that the indices need not be ordered, nor contigous, but the returned
    indices_concat will be a list of slices, and thus contiguous.

    Example
    -------

    >>> A = np.random.rand(100)
    >>> value = 0
    >>> new_value = 0
    >>>
    >>> indices_list = [slice(10, 12), slice(12, 19)]
    >>> for indices in indices_list:
    >>>     value += np.sum(A[indices])
    >>>
    >>> indices_concat, new_indices_list = concatenate_indices(indices_list)
    >>> new_value = np.sum(A[indices_concat])
    >>>
    >>> assert abs(value - new_value) < 1e-10
    >>>
    >>> B = A[indices_concat]
    >>> assert B.shape == (9, )
    >>> new_value = 0
    >>> for indices in new_indices_list:
    >>>     new_value += np.sum(B[indices])
    >>>
    >>> assert abs(value - new_value) < 1e-10

    Returns
    -------
        ``(indices_concat, new_indices_list)``
    """
    indices_list = list(indices_list)
    if len(indices_list) == 0:
        return slice(0), []  # type: ignore

    if not isinstance(indices_list[0], tuple):
        # If indices are not tuples, then wrap everything in tuples and recurse
        assert all([not isinstance(indices, tuple) for indices in indices_list])
        _indices_concat, _new_indices_list = _concatenate_indices([(indices, ) for indices in indices_list])
        return _indices_concat[0], [indices[0] for indices in _new_indices_list]

    # All indices are wrapped in tuples
    assert all([isinstance(indices, tuple) for indices in indices_list])
    return _concatenate_indices(indices_list)  # type: ignore


def _concatenate_indices(indices_list: Iterable[tuple[slice, ...]],
                         ) -> tuple[tuple[slice, ...], list[tuple[slice, ...]]]:
    """ See :func:`concatenate_indices`
    """
    limits_jis = np.array([[(index.start, index.stop, index.step) for index in indices]
                           for indices in indices_list])

    start_i = np.min(limits_jis[..., 0], axis=0)
    stop_i = np.max(limits_jis[..., 1], axis=0)

    indices_concat = tuple([slice(start, stop) for start, stop in zip(start_i, stop_i)])
    new_indices_list = [tuple([slice(start - startcat, stop - startcat, step)
                               for (startcat, (start, stop, step)) in zip(start_i, limits_is)])
                        for limits_is in limits_jis]

    return indices_concat, new_indices_list


def parulmopen(fname: str, comm: Communicator, *args, **kwargs):
    if comm.rank == 0:
        return open(fname, *args, **kwargs)
    else:
        return nullcontext()


def proxy_sknX_slicen(reader, *args, comm: Communicator) -> NDArray[np.complex128]:
    if len(args) == 0:
        A_sknX = reader
    else:
        A_sknX = reader.proxy(*args)
    nn = A_sknX.shape[2]
    nlocaln = (nn + comm.size - 1) // comm.size
    myslicen = slice(comm.rank * nlocaln, (comm.rank + 1) * nlocaln)
    my_A_sknX = np.array([[A_nX[myslicen] for A_nX in A_knX] for A_knX in A_sknX])

    return my_A_sknX


def add_fake_kpts(ksd: KohnShamDecomposition):
    """This function is necessary to read some fields without having a
    calculator attached.
    """

    class FakeKpt(NamedTuple):
        s: int
        k: int

    class FakeKsl(NamedTuple):
        using_blacs: bool = False

    # Figure out
    ksdreader = ksd.reader
    skshape = ksdreader.eig_un.shape[:2]
    kpt_u = [FakeKpt(s=s, k=k)
             for s in range(skshape[0])
             for k in range(skshape[1])]
    ksd.kpt_u = kpt_u
    ksd.ksl = FakeKsl()


def create_pulse(frequency: float,
                 fwhm: float = 5.0,
                 t0: float = 10.0,
                 print: Callable | None = None) -> GaussianPulse:
    """ Create Gaussian laser pulse.

    frequency
        Pulse frequncy in units of eV.
    fwhm
        Full width at half maximum in time domain in units of fs.
    t0
        Maximum of pulse envelope in units of fs.
    print
        Printing function to control verbosity.
    """
    if print is None:
        print = parprint

    # Pulse
    fwhm_eV = 8 * np.log(2) / (fwhm * fs_to_au) * au_to_eV
    tau = fwhm / (2 * np.sqrt(2 * np.log(2)))
    sigma = 1 / (tau * fs_to_au) * au_to_eV  # eV
    strength = 1e-6
    t0 = t0 * 1e3
    sincos = 'cos'
    print(f'Creating pulse at {frequency:.3f}eV with FWHM {fwhm:.2f}fs '
          f'({fwhm_eV:.2f}eV) t0 {t0:.1f}fs', flush=True)

    return GaussianPulse(strength, t0, frequency, sigma, sincos)


def get_gaussian_pulse_values(pulse: Perturbation) -> dict[str, float]:
    """ Get pulse frequency and FWHM of Gaussian pulse.

    Returns
    -------
    Empty dictionary if pulse is not `GaussianPulse`, otherwise dictionary with keys:

        ``pulsefreq`` - pulse frequency in units of eV.
        ``pulsefwhm`` - pulse full width at half-maximum in time domain in units of fs.
    """
    from gpaw.tddft.units import eV_to_au, au_to_fs

    d = pulse.todict()
    ret: dict[str, float] = dict()
    if d['name'] == 'GaussianPulse':
        ret['pulsefreq'] = d['frequency']
        ret['pulsefwhm'] = (1 / (d['sigma'] * eV_to_au) * au_to_fs *
                            (2 * np.sqrt(2 * np.log(2))))
    return ret


fast_pad_nice_factors = np.array([4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 8096])


def fast_pad(nt: int) -> int:
    """ Return a length that is at least twice as large as the given input,
    and the FFT of data of such length is fast.
    """
    padnt = 2 * nt
    insert = np.searchsorted(fast_pad_nice_factors, padnt)
    if insert <= len(fast_pad_nice_factors):
        padnt = fast_pad_nice_factors[insert]
    assert padnt >= 2 * nt
    return padnt


def format_string_to_glob(fmt: str) -> str:
    """ Convert a format string to a glob-type expression.

    Replaces all the replacement fields ``{...}`` in the format string
    with a glob ``*``.

    Example
    -------
    >>> format_string_to_glob('pulserho_pf{pulsefreq:.2f}/t{time:09.1f}{tag}.npy')
    pulserho_pf*/t*.npy

    Parameters
    ---------
    fmt
        Format string.

    Returns
    -------
    Glob-type expression.
    """
    # Replace replacement fields by *
    # Note how several replacement fields next to each other are
    # replaced by only one * thanks to the (...)+
    glob_expr = re.sub(r'({[^{}]*})+', '*', fmt)
    return glob_expr


def format_string_to_regex(fmt: str) -> re.Pattern:
    r""" Convert a format string to a regex expression.

    Replaces all the replacement fields ``{...}`` in the format string
    with a regular expression and escapes all special characters outside
    the replacement fields.

    Replacement fields for variables ``time``, ``freq``, ``pulsefreq``
    and ``pulsefwhm`` are replaced by regex matching floating point numbers.
    Replacement fields for variables ``reim`` and ``tag`` are replaced by
    regex matching alphabetic characters and dashes.
    Remaining replacement fields are replaced by regex matching alphabetic
    characters.

    This can be used to parse a formatted string in order to get back the original
    values.

    Example
    -------
    >>> fmt = 'pulserho_pf{pulsefreq:.2f}/t{time:09.1f}{tag}.npy'
    >>> s = fmt.format(pulsefreq=3.8, time=30000, tag='-Iomega')
    pulserho_pf3.80/t0030000.0-Iomega.npy
    >>> regex = format_string_to_regex(fmt)
    re.compile('pulserho_pf(?P<pulsefreq>[-+]?[\d.]+)/t(?P<time>[-+]?[\d.]+)(?P<tag>[-A-za-z]*)\.npy')
    >>> regex.fullmatch(s).groupdict()
    {'pulsefreq': '3.80', 'time': '0030000.0', 'tag': '-Iomega'}

    Parameters
    ---------
    fmt
        Format string.

    Returns
    -------
    Compiled regex pattern.

    Notes
    -----
    Replacement fields should be named and not contain any attributes or indexing.
    """
    regex_expr = str(fmt)

    # Split the expression by parts separated by replacement fields
    split = re.split(r'({[^{}]+})', regex_expr)
    # Every other element is guaranteed to be a replacement field
    # Escape everything that is not a replacement field
    split[::2] = [re.escape(s) for s in split[::2]]
    # Join the expression back together
    regex_expr = ''.join(split)

    # Replace float variables
    regex_expr = re.sub(r'{(time|freq|pulsefreq|pulsefwhm)[-:.\w]+}',
                        r'(?P<\1>[-+]?[\\d.]+)',
                        regex_expr)

    # Replace reim and tag
    regex_expr = re.sub(r'{(reim)}', r'(?P<\1>[A-za-z]+)', regex_expr)
    regex_expr = re.sub(r'{(tag)}', r'(?P<\1>[-A-za-z]*)', regex_expr)

    # Replace other
    regex_expr = re.sub(r'{(\w*)}', r'(?P<\1>[A-za-z]*)', regex_expr)

    compiled = re.compile(regex_expr)

    return compiled


def partial_format(fmt, **kwargs) -> str:
    """ Partially format the format string.

    Equivalent to calling ``fmt.format(**kwargs)`` but replacement fields
    that are not present in the ``**kwargs`` will be left in the format string.

    Parameters
    ----------
    fmt
        Format string.
    **kwargs
        Passed to the :py:meth:`str.format` call.

    Returns
    -------
    Partially formatted string.

    Example
    -------
    >>> fmt = 'pulserho_pf{pulsefreq:.2f}/t{time:09.1f}{tag}.npy'
    >>> partial_format(fmt, pulsefreq=3.8)
    pulserho_pf3.80/t{time:09.1f}{tag}.npy
    """
    def partial_format_single(s):
        try:
            # Try to format
            return s.format(**kwargs)
        except KeyError:
            # If the replacement field is not among the kwargs, return unchanged
            return s

    # Split the expression by parts separated by replacement fields
    split = re.split(r'({[^{}]+})', fmt)
    # Every other element is guaranteed to be a replacement field
    # Try to format each field separately
    split[1::2] = [partial_format_single(s) for s in split[1::2]]
    # Join the expression back together
    fmt = ''.join(split)

    return fmt


def find_files(fmt: str,
               log: Logger | None = None,
               *,
               expected_keys: list[str]):
    """ Find files in file system matching the format string :attr:`fmt`.

    This function walks the file tree and looks for file names matching the
    format string :attr:`fmt`.

    Parameters
    ----------
    fmt
        Format string.
    log
        Optional logger object.
    expected_keys
        List of replacement fields ``{...}`` that are expected to be parsed from
        the file names. Unexpected fields raise :py:exc:`ValueError`.

    Returns
    -------
    Dictionary with keys, sorted by the parsed values matching :attr:`expected_keys`:

        ``filename`` - List of filenames found.
        **key** - List of parsed value for each key in :attr:`expected_keys`.

    Example
    -------
    >>> fmt = 'pulserho_pf3.80/t{time:09.1f}{tag}.npy'
    >>> find_files(fmt, expected_keys=['time', 'tag'])
    {'filename': ['pulserho_pf3.80/t0000010.0.npy',
                  'pulserho_pf3.80/t0000010.0-Iomega.npy',
                  'pulserho_pf3.80/t0000060.0.npy',
                  'pulserho_pf3.80/t0000060.0-Iomega.npy'],
     'time': [10.0, 10.0, 60.0, 60.0],
     'tag': ['', '-Iomega', '', '-Iomega']}
    """
    if log is None:
        log = NoLogger()

    # Find base (containing no format string replacement fields)
    non_format_parents = [parent for parent in Path(fmt).parents
                          if '{' not in parent.name]
    base = non_format_parents[0] if len(non_format_parents) > 0 else Path('.')
    log(str(base), who='Find files', rank=0)

    # Express the format string relative to the base
    rel_pulserho_fmt = str(Path(fmt).relative_to(base))
    log(rel_pulserho_fmt, who='Find files', rank=0)

    # Convert format specifier to glob, and to regex
    pulserho_glob = format_string_to_glob(rel_pulserho_fmt)
    pulserho_regex = format_string_to_regex(rel_pulserho_fmt)
    log(pulserho_glob, who='Find files', rank=0)
    log(pulserho_regex.pattern, who='Find files', rank=0)

    matching = base.glob(pulserho_glob)

    found: list[dict[str, Any]] = []

    # Loop over the matching files
    for match in matching:
        relmatch = match.relative_to(base)
        m = pulserho_regex.fullmatch(str(relmatch))
        if m is None:
            continue
        metadata = {key: float(value) if key not in ['tag', 'reim'] else value
                    for key, value in m.groupdict().items()}
        fname = str(base / relmatch)
        test = fmt.format(**metadata)
        assert fname == test, fname + ' != ' + test
        if set(metadata.keys()) > set(expected_keys):
            raise ValueError(f'Found unexpected key in file name {base / relmatch}:\n'
                             f'Found {metadata}\nExpected {expected_keys}')
        log(relmatch, metadata, who='Find files', rank=0)
        metadata['filename'] = fname
        found.append(metadata)

    # Sort list of found files by expected_keys
    found = sorted(found, key=itemgetter(*expected_keys))

    # Unwrap so that we return one dictionary of lists
    ret = {key: [f.get(key, None) for f in found]
           for key in ['filename'] + expected_keys}

    return ret
